#!/usr/bin/env python

import argparse
import os
import subprocess
import sys

parser = argparse.ArgumentParser(description="Run the test suite, or some tests")
parser.add_argument("--coverage", "-c", action="store_true", help="Run with 'coverage'")
parser.add_argument(
    "--coverage-parallel", action="store_true", help="Run coverage with --parallel"
)
parser.add_argument("--verbose", "-v", action="store_true")
parser.add_argument(
    "--fast", "-f", action="store_true", help="Fast test run, skip end-to-end tests"
)
parser.add_argument(
    "--no-network",
    "-n",
    action="store_true",
    help="Skip network access (requires stuff has been fetched in previous run)",
)
parser.add_argument(
    "--show-browser", action="store_true", help="Don't hide web browser"
)
parser.add_argument(
    "--tests-root", action="store", default=".", help="Directory that holds 'tests' dir"
)
parser.add_argument(
    "test", type=str, nargs="*", help="Dotted path to a test module, case or method"
)

args = parser.parse_args()

cmd = ["-m", "unittest"]

if args.test:
    cmd.extend(args.test)
else:
    cmd.extend(
        [
            "discover",
            "-t",
            os.path.normpath(os.path.abspath(args.tests_root)),
            "-s",
            "tests",
        ]
    )

if args.verbose:
    cmd.append("-v")

if args.fast:
    os.environ["TEST_FAST_RUN"] = "1"

if args.show_browser:
    os.environ["TEST_SHOW_BROWSER"] = "1"

if args.no_network:
    os.environ["TEST_NO_NETWORK"] = "1"

if args.coverage:
    if args.coverage_parallel:
        cmd = ["-m", "coverage", "run", "--parallel"] + cmd
    else:
        cmd = ["-m", "coverage", "run"] + cmd

cmd.insert(0, "python")

sys.exit(subprocess.call(cmd))



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals

import unittest

from fluent.syntax import ast

from elm_fluent import exceptions
from elm_fluent.compiler import (
    compile_messages,
    message_function_name_for_msg_id,
    span_to_position,
)
from elm_fluent.stubs import defaults as dtypes, fluent

from .test_codegen import normalize_elm
from .utils import dedent_ftl


def compile_messages_to_elm(
    source,
    locale,
    module_name=None,
    include_module_line=False,
    include_imports=False,
    use_isolating=False,
    dynamic_html_attributes=True,
):
    module, errors, _ = compile_messages(
        dedent_ftl(source),
        locale,
        module_name=module_name,
        use_isolating=use_isolating,
        dynamic_html_attributes=dynamic_html_attributes,
    )
    return (
        module.as_source_code(
            include_module_line=include_module_line, include_imports=include_imports
        ),
        errors,
    )


class TestCompiler(unittest.TestCase):
    locale = "en-US"

    maxDiff = None

    def assertCodeEqual(self, code1, code2):
        self.assertEqual(normalize_elm(code2), normalize_elm(code1))

    def test_message_function_name_for_msg_id(self):
        self.assertEqual(message_function_name_for_msg_id("hello"), "hello")
        self.assertEqual(message_function_name_for_msg_id("hello-there"), "helloThere")
        self.assertEqual(message_function_name_for_msg_id("helloThere"), "helloThere")
        self.assertEqual(message_function_name_for_msg_id("hello.foo"), "hello_foo")
        self.assertEqual(
            message_function_name_for_msg_id("hello-html.foo"), "helloHtml_foo"
        )

    def test_single_string_literal(self):
        code, errs = compile_messages_to_elm(
            """
            foo = Foo ☺
        """,
            self.locale,
        )
        self.assertCodeEqual(
            code,
            """
            foo : Locale.Locale -> a -> String
            foo locale_ args_ =
                "Foo ☺"
        """,
        )
        self.assertEqual(errs, [])

    def test_module_line(self):
        code, errs = compile_messages_to_elm(
            """
            foo = Foo
        """,
            self.locale,
            module_name="Foo.Bar",
            include_module_line=True,
        )
        self.assertCodeEqual(
            code,
            """
            module Foo.Bar exposing (foo)

            foo : Locale.Locale -> a -> String
            foo locale_ args_ =
                "Foo"
        """,
        )
        self.assertEqual(errs, [])

    def test_string_literal_in_placeable(self):
        code, errs = compile_messages_to_elm(
            """
            foo = { "Foo" }
        """,
            self.locale,
        )
        self.assertCodeEqual(
            code,
            """
            foo : Locale.Locale -> a -> String
            foo locale_ args_ =
                "Foo"
        """,
        )
        self.assertEqual(errs, [])

    def test_number_literal(self):
        code, errs = compile_messages_to_elm(
            """
            foo = { 123 }
        """,
            self.locale,
        )
        self.assertCodeEqual(
            code,
            """
            foo : Locale.Locale -> a -> String
            foo locale_ args_ =
                NumberFormat.format (NumberFormat.fromLocale locale_) 123
        """,
        )
        self.assertEqual(errs, [])

    def test_inferred_number(self):
        # From the second instance of $count we know it is numeric,
        # so it must be treated as numeric in the first
        code, errs = compile_messages_to_elm(
            """
            foo = { $count }, { NUMBER($count) }
        """,
            self.locale,
        )
        self.assertCodeEqual(
            code,
            """
            foo : Locale.Locale -> { a | count : Fluent.FluentNumber number } -> String
            foo locale_ args_ =
                String.concat [ Fluent.formatNumber locale_ args_.count
                              , ", "
                              , Fluent.formatNumber locale_ args_.count
                              ]
        """,
        )
        self.assertEqual(errs, [])

    def test_inferred_number_from_select(self):
        code, errs = compile_messages_to_elm(
            """
            foo = { $count ->
               [one]   You have one item
              *[other] You have { $count } items
             }
        """,
            self.locale,
        )
        self.assertCodeEqual(
            code,
            """
            foo : Locale.Locale -> { a | count : Fluent.FluentNumber number } -> String
            foo locale_ args_ =
                case PluralRules.select (PluralRules.fromLocale locale_) (Fluent.numberValue args_.count) of
                    "one" ->
                        "You have one item"
                    _ ->
                        String.concat [ "You have "
                                      , Fluent.formatNumber locale_ args_.count
                                      , " items"
                                      ]
        """,
        )
        self.assertEqual(errs, [])

    def test_literal_number_in_select(self):
        code, errs = compile_messages_to_elm(
            """
            foo = { 123 ->
               [one]   You have one item
              *[other] You have more than one item
             }
        """,
            self.locale,
        )
        self.assertCodeEqual(
            code,
            """
            foo : Locale.Locale -> a -> String
            foo locale_ args_ =
                case PluralRules.select (PluralRules.fromLocale locale_) 123 of
                    "one" ->
                        "You have one item"
                    _ ->
                        "You have more than one item"
        """,
        )
        self.assertEqual(errs, [])

    def test_inferred_number_from_call(self):
        code, errs = compile_messages_to_elm(
            """
           bar = { NUMBER($count) }

           foo = { $count } - { bar }
        """,
            self.locale,
        )
        self.assertCodeEqual(
            code,
            """
            bar : Locale.Locale -> { a | count : Fluent.FluentNumber number } -> String
            bar locale_ args_ =
                Fluent.formatNumber locale_ args_.count

            foo : Locale.Locale -> { a | count : Fluent.FluentNumber number } -> String
            foo locale_ args_ =
                String.concat [ Fluent.formatNumber locale_ args_.count
                              , " - "
                              , bar locale_ args_
                              ]
        """,
        )
        self.assertEqual(errs, [])

    def test_inferred_number_from_call_reversed(self):
        code, errs = compile_messages_to_elm(
            """
            foo = { bar }

            bar = { $count } - { baz }

            baz = { NUMBER($count) }
        """,
            self.locale,
        )
        self.assertCodeEqual(
            code,
            """
            foo : Locale.Locale -> { a | count : Fluent.FluentNumber number } -> String
            foo locale_ args_ =
                bar locale_ args_

            bar : Locale.Locale -> { a | count : Fluent.FluentNumber number } -> String
            bar locale_ args_ =
                String.concat [ Fluent.formatNumber locale_ args_.count
                              , " - "
                              , baz locale_ args_
                              ]

            baz : Locale.Locale -> { a | count : Fluent.FluentNumber number } -> String
            baz locale_ args_ =
                Fluent.formatNumber locale_ args_.count
        """,
        )
        self.assertEqual(errs, [])

    def test_message_reference_plus_string_literal(self):
        code, errs = compile_messages_to_elm(
            """
            foo = Foo
            bar = X { foo }
        """,
            self.locale,
        )
        self.assertCodeEqual(
            code,
            """
            foo : Locale.Locale -> a -> String
            foo locale_ args_ =
                "Foo"

            bar : Locale.Locale -> a -> String
            bar locale_ args_ =
                String.concat [ "X "
                              , foo locale_ args_
                              ]
        """,
        )
        self.assertEqual(errs, [])

    def test_single_message_reference(self):
        code, errs = compile_messages_to_elm(
            """
            foo = Foo
            bar = { foo }
        """,
            self.locale,
        )
        self.assertCodeEqual(
            code,
            """
            foo : Locale.Locale -> a -> String
            foo locale_ args_ =
                "Foo"

            bar : Locale.Locale -> a -> String
            bar locale_ args_ =
                foo locale_ args_
        """,
        )
        self.assertEqual(errs, [])

    def test_message_attr_reference(self):
        code, errs = compile_messages_to_elm(
            """
            foo
               .attr = Foo Attr
            bar = { foo.attr }
        """,
            self.locale,
        )
        self.assertCodeEqual(
            code,
            """
            foo_attr : Locale.Locale -> a -> String
            foo_attr locale_ args_ =
                "Foo Attr"

            bar : Locale.Locale -> a -> String
            bar locale_ args_ =
                foo_attr locale_ args_
        """,
        )
        self.assertEqual(errs, [])

    def test_missing_attr_reference(self):
        code, errs = compile_messages_to_elm(
            """
            foo = Hello
               .attr = Foo Attr
            bar = { foo.baz }
        """,
            self.locale,
        )
        self.assertEqual(errs[0].error_sources[0].message_id, "bar")
        self.assertEqual(errs[0], exceptions.ReferenceError("Unknown message: foo.baz"))

    def test_single_message_reference_reversed_order(self):
        # We should cope with forward references
        code, errs = compile_messages_to_elm(
            """
            bar = { foo }
            foo = Foo
        """,
            self.locale,
        )
        self.assertCodeEqual(
            code,
            """
            bar : Locale.Locale -> a -> String
            bar locale_ args_ =
                foo locale_ args_

            foo : Locale.Locale -> a -> String
            foo locale_ args_ =
                "Foo"
        """,
        )
        self.assertEqual(errs, [])

    def test_single_message_bad_reference(self):
        code, errs = compile_messages_to_elm(
            """
            bar = { foo }
        """,
            self.locale,
        )
        self.assertIn("COMPILATION_ERROR", code)
        self.assertEqual(len(errs), 1)
        self.assertEqual(errs[0].error_sources[0].message_id, "bar")
        self.assertEqual(errs[0], exceptions.ReferenceError("Unknown message: foo"))

    def test_bad_name_keyword(self):
        code, errs = compile_messages_to_elm(
            """
            type = My String
        """,
            self.locale,
        )
        self.assertIn("COMPILATION_ERROR", code)
        self.assertEqual(len(errs), 1)
        self.assertEqual(errs[0].error_sources[0].message_id, "type")
        self.assertEqual(
            errs[0],
            exceptions.BadMessageId(
                "'type' is not allowed as a message ID because it clashes "
                "with an Elm keyword. Please choose another ID."
            ),
        )

    def test_bad_name_default_import(self):
        code, errs = compile_messages_to_elm(
            """
            max = My String
        """,
            self.locale,
        )
        self.assertIn("COMPILATION_ERROR", code)
        self.assertEqual(len(errs), 1)
        self.assertEqual(errs[0].error_sources[0].message_id, "max")
        self.assertEqual(
            errs[0],
            exceptions.BadMessageId(
                "'max' is not allowed as a message ID because it clashes "
                "with an Elm default import. Please choose another ID."
            ),
        )

    def test_bad_name_duplicate(self):
        code, errs = compile_messages_to_elm(
            """
            a-message-id = My Message

            aMessageId = Another Message
        """,
            self.locale,
        )
        self.assertIn("COMPILATION_ERROR", code)
        self.assertEqual(len(errs), 1)
        self.assertEqual(errs[0].error_sources[0].message_id, "aMessageId")
        self.assertEqual(
            errs[0],
            exceptions.BadMessageId(
                "'aMessageId' is not allowed as a message ID because it clashes "
                "with another message ID - 'a-message-id'. "
                "Please choose another ID."
            ),
        )

    def test_message_mapping_used(self):
        # Checking that we actually use message_mapping when looking up the name
        # of the message function to call.
        code, errs = compile_messages_to_elm(
            """
            a-message-id = Foo
        """,
            self.locale,
        )
        self.assertCodeEqual(
            code,
            """
            aMessageId : Locale.Locale -> a -> String
            aMessageId locale_ args_ =
                "Foo"
        """,
        )
        self.assertEqual(errs, [])

    def test_external_argument(self):
        code, errs = compile_messages_to_elm(
            """
            with-arg = Some text { $arg }
        """,
            self.locale,
        )
        self.assertCodeEqual(
            code,
            """
            withArg : Locale.Locale -> { a | arg : String } -> String
            withArg locale_ args_ =
                String.concat [ "Some text "
                              , args_.arg
                              ]
        """,
        )
        self.assertEqual(errs, [])

    def test_lone_external_argument(self):
        code, errs = compile_messages_to_elm(
            """
            with-arg = { $arg }
        """,
            self.locale,
        )
        self.assertCodeEqual(
            code,
            """
            withArg : Locale.Locale -> { a | arg : String } -> String
            withArg locale_ args_ =
                args_.arg
        """,
        )
        self.assertEqual(errs, [])

    def test_function_call(self):
        code, errs = compile_messages_to_elm(
            """
                 foo = { NUMBER(1234) }
        """,
            self.locale,
        )
        self.assertEqual(errs, [])
        self.assertCodeEqual(
            code,
            """
            foo : Locale.Locale -> a -> String
            foo locale_ args_ =
                NumberFormat.format (NumberFormat.fromLocale locale_) 1234
        """,
        )

    def test_function_call_external_arg(self):
        code, errs = compile_messages_to_elm(
            """
            foo = { NUMBER($arg) }
        """,
            self.locale,
        )
        self.assertCodeEqual(
            code,
            """
            foo : Locale.Locale -> { a | arg : Fluent.FluentNumber number } -> String
            foo locale_ args_ =
                Fluent.formatNumber locale_ args_.arg
        """,
        )
        self.assertEqual(errs, [])

    def test_NUMBER_useGrouping(self):
        code, errs = compile_messages_to_elm(
            """
            foo = { NUMBER($arg, useGrouping: 0) }
        """,
            self.locale,
        )
        self.assertCodeEqual(
            code,
            """
            foo : Locale.Locale -> { a | arg : Fluent.FluentNumber number } -> String
            foo locale_ args_ =
                let
                    initial_opts_ = Fluent.numberFormattingOptions args_.arg
                    fnum_ = Fluent.reformattedNumber { initial_opts_ | locale = locale_, useGrouping = False } args_.arg
                in
                    Fluent.formatNumber locale_ fnum_
        """,
        )
        self.assertEqual(errs, [])

    def test_NUMBER_minimumIntegerDigits(self):
        code, errs = compile_messages_to_elm(
            """
            foo = { NUMBER($arg, minimumIntegerDigits: 2) }
        """,
            self.locale,
        )
        self.assertCodeEqual(
            code,
            """
            foo : Locale.Locale -> { a | arg : Fluent.FluentNumber number } -> String
            foo locale_ args_ =
                let
                    initial_opts_ = Fluent.numberFormattingOptions args_.arg
                    fnum_ = Fluent.reformattedNumber { initial_opts_ | locale = locale_, minimumIntegerDigits = Just 2 } args_.arg
                in
                    Fluent.formatNumber locale_ fnum_
        """,
        )
        self.assertEqual(errs, [])

    def test_NUMBER_everything(self):
        code, errs = compile_messages_to_elm(
            """
            foo =
               There are { NUMBER(7890,
                                  useGrouping: 0,
                                  minimumIntegerDigits: 2,
                                  minimumFractionDigits: 3,
                                  maximumFractionDigits: 4,
                                  minimumSignificantDigits: 5,
                                  maximumSignificantDigits: 6 ) } things
        """,
            self.locale,
        )
        self.assertCodeEqual(
            code,
            """
            foo : Locale.Locale -> a -> String
            foo locale_ args_ =
                let
                    defaults_ = NumberFormat.defaults
                    fnum_ = Fluent.formattedNumber { defaults_ | locale = locale_, maximumFractionDigits = Just 4, maximumSignificantDigits = Just 6, minimumFractionDigits = Just 3, minimumIntegerDigits = Just 2, minimumSignificantDigits = Just 5, useGrouping = False } 7890
                in
                    String.concat [ "There are "
                                  , Fluent.formatNumber locale_ fnum_
                                  , " things"
                                  ]
        """,
        )
        self.assertEqual(errs, [])

    def test_inferred_date(self):
        # From the second instance of $date we know it is a date,
        # so it must be treated as a date in the first
        code, errs = compile_messages_to_elm(
            """
            foo = { $startdate }, { DATETIME($startdate) }
        """,
            self.locale,
        )
        self.assertCodeEqual(
            code,
            """
            foo : Locale.Locale -> { a | startdate : Fluent.FluentDate } -> String
            foo locale_ args_ =
                String.concat [ Fluent.formatDate locale_ args_.startdate
                              , ", "
                              , Fluent.formatDate locale_ args_.startdate
                              ]
        """,
        )
        self.assertEqual(errs, [])

    def test_DATETIME_everything(self):
        code, errs = compile_messages_to_elm(
            """
            foo = { DATETIME($startdate,
                             hour12: 0,
                             weekday: "short",
                             era: "long",
                             year: "numeric",
                             month: "numeric",
                             day: "numeric",
                             hour: "2-digit",
                             minute: "2-digit",
                             second: "2-digit",
                             timeZoneName: "short") }
        """,
            self.locale,
        )
        self.assertCodeEqual(
            code,
            """
            foo : Locale.Locale -> { a | startdate : Fluent.FluentDate } -> String
            foo locale_ args_ =
                let
                    initial_opts_ = Fluent.dateFormattingOptions args_.startdate
                    fdate_ = Fluent.reformattedDate { initial_opts_ | day = DateTimeFormat.NumericNumber, era = DateTimeFormat.LongName, hour = DateTimeFormat.TwoDigitNumber, hour12 = Just False, locale = locale_, minute = DateTimeFormat.TwoDigitNumber, month = DateTimeFormat.NumericMonth, second = DateTimeFormat.TwoDigitNumber, timeZoneName = DateTimeFormat.ShortTimeZone, weekday = DateTimeFormat.ShortName, year = DateTimeFormat.NumericNumber } args_.startdate
                in
                    Fluent.formatDate locale_ fdate_
        """,
        )
        self.assertEqual(errs, [])

    def test_missing_function_call(self):
        code, errs = compile_messages_to_elm(
            """
            foo = { MISSING(123) }
        """,
            self.locale,
        )
        self.assertEqual(errs[0].error_sources[0].message_id, "foo")
        self.assertEqual(
            errs[0], exceptions.ReferenceError("Unknown function: MISSING")
        )

    def test_function_call_with_bad_keyword_arg(self):
        code, errs = compile_messages_to_elm(
            """
            foo = { NUMBER(123, foo: 1) }
        """,
            self.locale,
        )
        self.assertEqual(len(errs), 1)
        self.assertEqual(errs[0].error_sources[0].message_id, "foo")
        self.assertEqual(
            errs[0],
            exceptions.FunctionParameterError(
                "NUMBER() got an unexpected keyword argument 'foo'"
            ),
        )

    def test_function_call_with_bad_positional_arg(self):
        code, errs = compile_messages_to_elm(
            """
            foo = { NUMBER(123, 456) }
        """,
            self.locale,
        )
        self.assertEqual(len(errs), 1)
        self.assertEqual(errs[0].error_sources[0].message_id, "foo")
        self.assertEqual(
            errs[0],
            exceptions.FunctionParameterError(
                "NUMBER() takes 1 positional argument(s) but 2 were given"
            ),
        )
        self.assertEqual(type(errs[0].error_sources[0].expr), ast.CallExpression)

    def test_message_arg_type_mismatch(self):
        # Should return error gracefully
        src = dedent_ftl(
            """
            foo = { NUMBER($arg) } { DATETIME($arg) }
        """
        )
        code, errs = compile_messages_to_elm(src, self.locale)
        self.assertEqual(len(errs), 1)
        err = errs[0]
        self.assertEqual(span_to_position(err.error_sources[0].expr.span, src), (1, 26))

        self.assertEqual(type(err), exceptions.RecordTypeMismatch)
        self.assertEqual(err.field_name, "arg")
        type_sources = err.record_type.field_type_ftl_sources["arg"]
        self.assertEqual(len(type_sources), 2)

        self.assertEqual(
            span_to_position(type_sources[0].ftl_source.expr.span, src), (1, 9)
        )
        self.assertEqual(type_sources[0].type_obj, fluent.FluentNumber)

        self.assertEqual(
            span_to_position(type_sources[1].ftl_source.expr.span, src), (1, 26)
        )
        self.assertEqual(type_sources[1].type_obj, fluent.FluentDate)

    def test_message_arg_type_mismatch_across_messsages(self):
        # Should return error gracefully, including info about where the
        # different types were inferred from
        src = dedent_ftl(
            """
            foo = { bar } { baz }

            bar = { NUMBER($arg) }

            baz = { DATETIME($arg) }
        """
        )
        code, errs = compile_messages_to_elm(src, self.locale)
        self.assertEqual(len(errs), 1)
        err = errs[0]
        self.assertEqual(span_to_position(err.error_sources[0].expr.span, src), (1, 17))

        self.assertEqual(type(err), exceptions.RecordTypeMismatch)
        self.assertEqual(err.field_name, "arg")
        type_sources = err.record_type.field_type_ftl_sources["arg"]
        self.assertEqual(len(type_sources), 2)
        self.assertEqual(
            span_to_position(type_sources[0].ftl_source.expr.span, src), (3, 9)
        )
        self.assertEqual(
            span_to_position(type_sources[1].ftl_source.expr.span, src), (5, 9)
        )

    def test_message_arg_type_mismatch_with_args(self):
        # Should return error gracefully
        code, errs = compile_messages_to_elm(
            """
            foo = { NUMBER($arg, useGrouping:0) } { DATETIME($arg, era:"long") }
        """,
            self.locale,
        )
        self.assertEqual(len(errs), 1)

    def test_message_arg_type_mismatch_with_string(self):
        src = dedent_ftl(
            """
            foo = { bar } { baz }

            bar = { $arg }

            baz = { NUMBER($arg) }
        """
        )
        code, errs = compile_messages_to_elm(src, self.locale)
        self.assertEqual(len(errs), 1)
        err = errs[0]
        self.assertEqual(span_to_position(err.error_sources[0].expr.span, src), (1, 17))

        self.assertEqual(type(err), exceptions.RecordTypeMismatch)
        self.assertEqual(err.field_name, "arg")
        type_sources = err.record_type.field_type_ftl_sources["arg"]
        self.assertEqual(len(type_sources), 2)
        self.assertEqual(
            span_to_position(type_sources[0].ftl_source.expr.span, src), (3, 7)
        )
        self.assertEqual(
            span_to_position(type_sources[1].ftl_source.expr.span, src), (5, 9)
        )

    def test_function_arg_type_mismatch(self):
        code, errs = compile_messages_to_elm(
            """
            foo = { NUMBER($arg, era: 123) }
        """,
            self.locale,
        )
        self.assertEqual(len(errs), 1)
        self.assertEqual(type(errs[0].error_sources[0].expr), ast.CallExpression)

    def test_message_with_attrs(self):
        code, errs = compile_messages_to_elm(
            """
            foo = Foo
               .attr-1 = Attr 1
               .attr-2 = Attr 2
        """,
            self.locale,
        )
        self.assertCodeEqual(
            code,
            """
            foo : Locale.Locale -> a -> String
            foo locale_ args_ =
                "Foo"

            foo_attr1 : Locale.Locale -> a -> String
            foo_attr1 locale_ args_ =
                "Attr 1"

            foo_attr2 : Locale.Locale -> a -> String
            foo_attr2 locale_ args_ =
                "Attr 2"
        """,
        )
        self.assertEqual(errs, [])

    def test_term_inline(self):
        code, errs = compile_messages_to_elm(
            """
           -term = Term
           message = Message { -term }
        """,
            self.locale,
        )
        self.assertEqual(errs, [])
        self.assertCodeEqual(
            code,
            """
            message : Locale.Locale -> a -> String
            message locale_ args_ =
                "Message Term"
        """,
        )

    def test_variant_select_inline(self):
        code, errs = compile_messages_to_elm(
            """
            -my-term = {
                [a] A
               *[b] B
              }
            foo = Before { -my-term[a] } After
        """,
            self.locale,
        )
        self.assertCodeEqual(
            code,
            """
            foo : Locale.Locale -> a -> String
            foo locale_ args_ =
                "Before A After"
        """,
        )
        self.assertEqual(errs, [])

    def test_variant_select_default(self):
        code, errs = compile_messages_to_elm(
            """
            -my-term = {
                [a] A
               *[b] B
              }
            foo = { -my-term }
        """,
            self.locale,
        )
        self.assertCodeEqual(
            code,
            """
            foo : Locale.Locale -> a -> String
            foo locale_ args_ =
                "B"
        """,
        )
        self.assertEqual(errs, [])

    def test_variant_select_missing_variant(self):
        # We don't use the default, it is more useful to give a compilation error
        code, errs = compile_messages_to_elm(
            """
            -my-term = {
                [a] A
               *[b] B
              }
            foo = { -my-term[c] }
        """,
            self.locale,
        )
        self.assertEqual(errs[0].error_sources[0].message_id, "foo")
        self.assertEqual(
            errs[0], exceptions.ReferenceError("Unknown variant: -my-term[c]")
        )

    def test_variant_select_missing_term(self):
        code, errs = compile_messages_to_elm(
            """
            foo = { -my-term[a] }
        """,
            self.locale,
        )
        self.assertEqual(errs[0].error_sources[0].message_id, "foo")
        self.assertEqual(errs[0], exceptions.ReferenceError("Unknown term: -my-term"))

    def test_variant_select_from_non_variant(self):
        code, errs = compile_messages_to_elm(
            """
            -my-term = Term
            foo = { -my-term[a] }
        """,
            self.locale,
        )
        self.assertEqual(errs[0].error_sources[0].message_id, "foo")
        self.assertEqual(
            errs[0], exceptions.ReferenceError("Unknown variant: -my-term[a]")
        )
        self.assertEqual(len(errs), 1)

    def test_select_string(self):
        code, errs = compile_messages_to_elm(
            """
            foo = { $stringArg ->
                [a] A
               *[b] B
             }
        """,
            self.locale,
        )
        self.assertCodeEqual(
            code,
            """
            foo : Locale.Locale -> { a | stringArg : String } -> String
            foo locale_ args_ =
                case args_.stringArg of
                    "a" ->
                        "A"
                    _ ->
                        "B"
        """,
        )
        self.assertEqual(errs, [])

    def test_select_number(self):
        code, errs = compile_messages_to_elm(
            """
            foo = { $numberArg ->
                [1] One
               *[2] { 2 }
             }
        """,
            self.locale,
        )
        # We should not get number formatting calls in the select expression or
        # or the key comparisons, but we should get them for the select value
        # for { 2 }.
        # We should also get $numberArg inferred to be numeric
        self.assertCodeEqual(
            code,
            """
            foo : Locale.Locale -> { a | numberArg : Fluent.FluentNumber number } -> String
            foo locale_ args_ =
                case Fluent.numberValue args_.numberArg of
                    1 ->
                        "One"
                    _ ->
                        NumberFormat.format (NumberFormat.fromLocale locale_) 2
        """,
        )
        self.assertEqual(errs, [])

    def test_select_number_literal(self):
        code, errs = compile_messages_to_elm(
            """
            foo = { 1 ->
                [1] One
               *[2] Two
             }
        """,
            self.locale,
        )
        self.assertCodeEqual(
            code,
            """
            foo : Locale.Locale -> a -> String
            foo locale_ args_ =
                "One"
            """,
        )
        self.assertEqual(errs, [])

    def test_select_plural_categories(self):
        code, errs = compile_messages_to_elm(
            """
            foo = { $count ->
                [one] You have one thing
               *[other] You have some things
             }
        """,
            self.locale,
        )
        self.assertCodeEqual(
            code,
            """
            foo : Locale.Locale -> { a | count : Fluent.FluentNumber number } -> String
            foo locale_ args_ =
                case PluralRules.select (PluralRules.fromLocale locale_) (Fluent.numberValue args_.count) of
                    "one" ->
                        "You have one thing"
                    _ ->
                        "You have some things"
        """,
        )

    def test_select_mixed_numeric(self):
        code, errs = compile_messages_to_elm(
            """
            foo = { $count ->
                [0] You have nothing
                [one] You have a thing
               *[other] You have some things
             }
        """,
            self.locale,
        )
        self.assertCodeEqual(
            code,
            """
            foo : Locale.Locale -> { a | count : Fluent.FluentNumber number } -> String
            foo locale_ args_ =
                let
                    val_ = Fluent.numberValue args_.count
                    pl_ = PluralRules.select (PluralRules.fromLocale locale_) val_
                in
                    if (val_ == 0) then
                        "You have nothing"
                    else
                        if (pl_ == "one") then
                            "You have a thing"
                        else
                            "You have some things"
        """,
        )

    def test_select_mismatch(self):
        src = dedent_ftl(
            """
            foo = { 1 ->
                [x]   X
               *[y]   Y
             }
        """
        )

        code, errs = compile_messages_to_elm(src, self.locale)
        self.assertEqual(len(errs), 1)
        err = errs[0]
        self.assertEqual(type(err), exceptions.TypeMismatch)
        self.assertEqual(span_to_position(err.error_sources[0].expr.span, src), (1, 9))
        # TODO - it would be nice to capture the other places that are causing
        # the type mismatch (e.g. the [x] key), but we don't have the
        # infrastructure for that and it is low priority.

    def test_select_mismtatch_with_arg(self):
        src = dedent_ftl(
            """
            foo = { NUMBER($count) ->
                [x]   X
               *[y]   Y
             }
        """
        )

        code, errs = compile_messages_to_elm(src, self.locale)
        self.assertEqual(len(errs), 1)
        err = errs[0]
        self.assertEqual(span_to_position(err.error_sources[0].expr.span, src), (1, 9))

        self.assertEqual(type(err), exceptions.RecordTypeMismatch)
        self.assertEqual(err.field_name, "count")
        type_sources = err.record_type.field_type_ftl_sources["count"]
        self.assertEqual(len(type_sources), 2)

        self.assertEqual(
            span_to_position(type_sources[0].ftl_source.expr.span, src), (1, 9)
        )
        self.assertEqual(type_sources[0].type_obj, fluent.FluentNumber)

        self.assertEqual(
            span_to_position(type_sources[1].ftl_source.expr.span, src), (2, 5)
        )
        self.assertEqual(type_sources[1].type_obj, dtypes.String)

    def test_select_mixed_numeric_last_not_default(self):
        code, errs = compile_messages_to_elm(
            """
            foo = { $count ->
                [0]     You have nothing
               *[other] You have some things
                [one]   You have a thing
             }
        """,
            self.locale,
        )
        self.assertCodeEqual(
            code,
            """
            foo : Locale.Locale -> { a | count : Fluent.FluentNumber number } -> String
            foo locale_ args_ =
                let
                    val_ = Fluent.numberValue args_.count
                    pl_ = PluralRules.select (PluralRules.fromLocale locale_) val_
                in
                    if (val_ == 0) then
                        "You have nothing"
                    else
                        if (pl_ == "other") then
                            "You have some things"
                        else
                            if (pl_ == "one") then
                                "You have a thing"
                            else
                                "You have some things"
        """,
        )

    def test_combine_strings(self):
        code, errs = compile_messages_to_elm(
            """
            foo = Start { "Middle" } End
        """,
            self.locale,
        )
        self.assertCodeEqual(
            code,
            """
            foo : Locale.Locale -> a -> String
            foo locale_ args_ =
                "Start Middle End"
        """,
        )
        self.assertEqual(errs, [])

    def test_single_string_literal_isolating(self):
        code, errs = compile_messages_to_elm(
            """
            foo = Foo
        """,
            self.locale,
            use_isolating=True,
        )
        # No isolating chars, because we have no placeables.
        self.assertCodeEqual(
            code,
            """
            foo : Locale.Locale -> a -> String
            foo locale_ args_ =
                "Foo"
        """,
        )
        self.assertEqual(errs, [])

    def test_interpolation_isolating(self):
        code, errs = compile_messages_to_elm(
            """
            foo = Foo { $arg } Bar
        """,
            self.locale,
            use_isolating=True,
        )
        self.assertCodeEqual(
            code,
            """
            foo : Locale.Locale -> { a | arg : String } -> String
            foo locale_ args_ =
                String.concat [ "Foo \u2068"
                              , args_.arg
                              , "\u2069 Bar"
                              ]
        """,
        )
        self.assertEqual(errs, [])

    def test_cycle_detection(self):
        code, errs = compile_messages_to_elm(
            """
            foo = { foo }
        """,
            self.locale,
        )
        self.assertEqual(errs[0].error_sources[0].message_id, "foo")
        self.assertEqual(
            errs[0], exceptions.CyclicReferenceError("Cyclic reference in foo")
        )

    def test_cycle_detection_with_attrs(self):
        code, errs = compile_messages_to_elm(
            """
            foo
               .attr1 = { bar.attr2 }

            bar
               .attr2 = { foo.attr1 }
        """,
            self.locale,
        )
        self.assertEqual(errs[0].error_sources[0].message_id, "foo.attr1")
        self.assertEqual(
            errs[0], exceptions.CyclicReferenceError("Cyclic reference in foo.attr1")
        )
        self.assertEqual(errs[1].error_sources[0].message_id, "bar.attr2")
        self.assertEqual(
            errs[1], exceptions.CyclicReferenceError("Cyclic reference in bar.attr2")
        )

    def test_term_cycle_detection(self):
        code, errs = compile_messages_to_elm(
            """
            -cyclic-term = { -cyclic-term }
            cyclic-term-message = { -cyclic-term }
        """,
            self.locale,
        )
        self.assertEqual(errs[0].error_sources[0].message_id, "cyclic-term-message")
        self.assertEqual(
            errs[0],
            exceptions.CyclicReferenceError("Cyclic reference in cyclic-term-message"),
        )

    def test_multiline_text(self):
        code, errs = compile_messages_to_elm(
            """
            test = Some text
                   that spans multiple lines
            """,
            self.locale,
        )
        self.assertCodeEqual(
            code,
            """
            test : Locale.Locale -> a -> String
            test locale_ args_ =
                "Some text\\nthat spans multiple lines"
            """,
        )

    def test_imports_eliminated(self):
        code, errs = compile_messages_to_elm(
            """
            test = Some text
            """,
            self.locale,
            include_imports=True,
        )
        self.assertCodeEqual(
            code,
            """
            import Intl.Locale as Locale

            test : Locale.Locale -> a -> String
            test locale_ args_ =
                "Some text"
            """,
        )

    def test_parameterized_terms_strings(self):
        code, errs = compile_messages_to_elm(
            """
            -thing = { $article ->
                  *[definite] the thing
                   [indefinite] a thing
                   [none] thing
            }
            thing-no-arg = { -thing }
            thing-no-arg-alt = { -thing() }
            thing-with-arg = { -thing(article: "indefinite") }
            thing-fallback = { -thing(article: "somethingelse") }
            """,
            self.locale,
        )
        self.assertCodeEqual(
            code,
            """
            thingNoArg : Locale.Locale -> a -> String
            thingNoArg locale_ args_ =
                "the thing"

            thingNoArgAlt : Locale.Locale -> a -> String
            thingNoArgAlt locale_ args_ =
                "the thing"

            thingWithArg : Locale.Locale -> a -> String
            thingWithArg locale_ args_ =
                "a thing"

            thingFallback : Locale.Locale -> a -> String
            thingFallback locale_ args_ =
                "the thing"
            """,
        )
        self.assertEqual(errs, [])

    def test_parameterized_terms_numbers(self):
        code, errs = compile_messages_to_elm(
            """
            -thing = { $count ->
                  *[1] one thing
                   [2] two things
            }
            thing-no-arg = { -thing }
            thing-no-arg-alt = { -thing() }
            thing-one = { -thing(count: 1) }
            thing-two = { -thing(count: 2) }
            """,
            self.locale,
        )
        self.assertCodeEqual(
            code,
            """
            thingNoArg : Locale.Locale -> a -> String
            thingNoArg locale_ args_ =
                "one thing"

            thingNoArgAlt : Locale.Locale -> a -> String
            thingNoArgAlt locale_ args_ =
                "one thing"

            thingOne : Locale.Locale -> a -> String
            thingOne locale_ args_ =
                "one thing"

            thingTwo : Locale.Locale -> a -> String
            thingTwo locale_ args_ =
                "two things"
            """,
        )

    def test_parameterized_terms_missing(self):
        code, errs = compile_messages_to_elm(
            """
            bad-term = { -missing(foo: "bar") }
            """,
            self.locale,
        )
        self.assertEqual(errs, [exceptions.ReferenceError("Unknown term: -missing")])
        self.assertEqual(errs[0].error_sources[0].message_id, "bad-term")

    def test_parameterized_terms_positional_arg(self):
        code, errs = compile_messages_to_elm(
            """
            -thing = { $article ->
                  *[definite] the thing
                   [indefinite] a thing
            }
            thing-positional-arg = { -thing("bar") }
            """,
            self.locale,
        )
        self.assertEqual(
            errs,
            [
                exceptions.TermParameterError(
                    "Positional arguments passed to term '-thing'"
                )
            ],
        )
        self.assertEqual(errs[0].error_sources[0].message_id, "thing-positional-arg")

    def test_parameterized_term_attributes(self):
        code, errs = compile_messages_to_elm(
            """
            -brand = Cool Thing
                .status = { $version ->
                    [v2]     available
                   *[v1]     deprecated
                }

            attr-with-arg = { -brand } is { -brand.status(version: "v2") ->
                 [available]   available, yay!
                *[deprecated]  deprecated, sorry
            }
            """,
            self.locale,
        )
        self.assertCodeEqual(
            code,
            """
            attrWithArg : Locale.Locale -> a -> String
            attrWithArg locale_ args_ =
                "Cool Thing is available, yay!"
            """,
        )

    def test_superfluous_or_bad_parameters(self):
        code, errs = compile_messages_to_elm(
            """
            -term = My Thing
            message = { -term(foo: "bar") }
            -parameterized-term = { $foo ->
                    *[a]   A
                     [b]   { -other-term }
             }
            -other-term = { $bar ->
                    *[c]   C
                     [d]   D
             }
            message-2 = { -parameterized-term(fooo: "bar") }
            """,
            self.locale,
        )
        self.assertEqual(
            errs,
            [
                exceptions.TermParameterError(
                    "Parameter 'foo' was passed to term '-term' which does not take parameters."
                ),
                exceptions.TermParameterError(
                    "Parameter 'fooo' was passed to term '-parameterized-term' which does not take this parameter. Did you mean: bar, foo?"
                ),
            ],
        )

    def test_messages_called_from_terms(self):
        code, errs = compile_messages_to_elm(
            """
            msg = Msg is { NUMBER($arg) }
            -foo = { $arg } { msg }
            ref-foo = { -foo(arg: 1) }
            """,
            self.locale,
        )
        # This construct is technically allowed at the moment, but might be
        # disallowed in future, and it doesn't make a huge amount of sense,
        # so we disallow for now.
        self.assertEqual(
            errs, [exceptions.ReferenceError("Message 'msg' called from within a term")]
        )
        self.assertEqual(errs[0].error_sources[0].message_id, "ref-foo")


class TestHtml(unittest.TestCase):
    locale = "en-US"

    maxDiff = None

    def assertCodeEqual(self, code1, code2):
        self.assertEqual(normalize_elm(code2), normalize_elm(code1))

    def test_text(self):
        code, errs = compile_messages_to_elm(
            """
            text-html = Me &amp; my friends ☺
            """,
            self.locale,
        )
        self.assertCodeEqual(
            code,
            """
            textHtml : Locale.Locale -> a -> List (String, List (Html.Attribute msg)) -> List (Html.Html msg)
            textHtml locale_ args_ attrs_ =
                [ Html.text "Me & my friends ☺"
                ]
            """,
        )
        self.assertEqual(errs, [])

    def test_tags(self):
        code, errs = compile_messages_to_elm(
            """
            tags-html = Some <b>bold text</b> and some <b>bold and <i>nested italic ☺</i></b> text
            """,
            self.locale,
            dynamic_html_attributes=False,
        )
        self.assertCodeEqual(
            code,
            """
            tagsHtml : Locale.Locale -> a -> List (String, List (Html.Attribute msg)) -> List (Html.Html msg)
            tagsHtml locale_ args_ attrs_ =
                [ Html.text "Some "
                , Html.b [] [ Html.text "bold text"
                            ]
                , Html.text " and some "
                , Html.b [] [ Html.text "bold and "
                            , Html.i [] [ Html.text "nested italic ☺"
                                        ]
                            ]
                , Html.text " text"
                ]
            """,
        )
        self.assertEqual(errs, [])

    def test_tags_not_builtin(self):
        code, errs = compile_messages_to_elm(
            """
            new-tag-html = <html5000newelement></html5000newelement>
            """,
            self.locale,
            dynamic_html_attributes=False,
        )
        self.assertCodeEqual(
            code,
            """
            newTagHtml : Locale.Locale -> a -> List (String, List (Html.Attribute msg)) -> List (Html.Html msg)
            newTagHtml locale_ args_ attrs_ =
                [ Html.node "html5000newelement" [] []
                ]
            """,
        )
        self.assertEqual(errs, [])

    def test_static_attributes(self):
        code, errs = compile_messages_to_elm(
            """
            tag-html = <b id="myid" data-foo data-bar="baz">text</b>
            """,
            self.locale,
            dynamic_html_attributes=False,
        )
        self.assertCodeEqual(
            code,
            """
            tagHtml : Locale.Locale -> a -> List (String, List (Html.Attribute msg)) -> List (Html.Html msg)
            tagHtml locale_ args_ attrs_ =
                [ Html.b [ Attributes.attribute "data-bar" "baz"
                         , Attributes.attribute "data-foo" ""
                         , Attributes.id "myid"
                         ] [ Html.text "text"
                           ]
                ]
            """,
        )
        self.assertEqual(errs, [])

    def test_class_attributes(self):
        # Check we work around bs4 'helpfulness'
        code, errs = compile_messages_to_elm(
            """
            new-tag-html = <b class="foo">text</b>
            """,
            self.locale,
            dynamic_html_attributes=False,
        )
        self.assertCodeEqual(
            code,
            """
            newTagHtml : Locale.Locale -> a -> List (String, List (Html.Attribute msg)) -> List (Html.Html msg)
            newTagHtml locale_ args_ attrs_ =
                [ Html.b [ Attributes.class "foo"
                         ] [ Html.text "text"
                           ]
                ]
            """,
        )
        self.assertEqual(errs, [])

    def test_dynamic_attributes(self):
        code, errs = compile_messages_to_elm(
            """
            attributes-html = <b class="foo" data-foo>text</b>
            """,
            self.locale,
        )
        self.assertCodeEqual(
            code,
            """
            attributesHtml : Locale.Locale -> a -> List (String, List (Html.Attribute msg)) -> List (Html.Html msg)
            attributesHtml locale_ args_ attrs_ =
                [ Html.b (List.concat [ [ Attributes.class "foo"
                                        , Attributes.attribute "data-foo" ""
                                        ]
                                      , Fluent.selectAttributes attrs_ [ "b"
                                                                       , ".foo"
                                                                       , "b.foo"
                                                                       , "[data-foo]"
                                                                       , "b[data-foo]"
                                                                       , "[data-foo=\\"\\"]"
                                                                       , "b[data-foo=\\"\\"]"
                                                                       ]
                                      ]) [ Html.text "text"
                                         ]
                ]
            """,
        )
        self.assertEqual(errs, [])

    def test_dynamic_attributes_2(self):
        code, errs = compile_messages_to_elm(
            """
            attributes-html = <b>text</b>
            """,
            self.locale,
        )
        self.assertCodeEqual(
            code,
            """
            attributesHtml : Locale.Locale -> a -> List (String, List (Html.Attribute msg)) -> List (Html.Html msg)
            attributesHtml locale_ args_ attrs_ =
                [ Html.b (Fluent.selectAttributes attrs_ [ "b"
                                                         ]) [ Html.text "text"
                                                            ]
                ]
            """,
        )
        self.assertEqual(errs, [])

    def test_attribute_substitution(self):
        # If we have any non-static text in attributes, we can't use them for attribute selectors
        code, errs = compile_messages_to_elm(
            """
            attributes-html = <b class="foo{ bar }" data-foo="{ $arg }" id="{ baz.id }">text</b>
            bar = Bar
            baz = baz
                .id = bazid
            """,
            self.locale,
        )
        self.assertCodeEqual(
            code,
            """
            attributesHtml : Locale.Locale -> { a | arg : String } -> List (String, List (Html.Attribute msg)) -> List (Html.Html msg)
            attributesHtml locale_ args_ attrs_ =
                [ Html.b (List.concat [ [ Attributes.class (String.concat [ "foo"
                                                                          , bar locale_ args_
                                                                          ])
                                        , Attributes.attribute "data-foo" args_.arg
                                        , Attributes.id (baz_id locale_ args_)
                                        ]
                                      , Fluent.selectAttributes attrs_ [ "b"
                                                                       , "[data-foo]"
                                                                       , "b[data-foo]"
                                                                       ]
                                      ]) [ Html.text "text"
                                         ]
                ]

            bar : Locale.Locale -> a -> String
            bar locale_ args_ =
                "Bar"

            baz : Locale.Locale -> a -> String
            baz locale_ args_ =
                "baz"

            baz_id : Locale.Locale -> a -> String
            baz_id locale_ args_ =
                "bazid"
            """,
        )
        self.assertEqual(errs, [])

    def test_non_string_args_in_attributes(self):
        code, errs = compile_messages_to_elm(
            """
            number-attribute-html = <b data-foo="{ NUMBER($arg) }">text</b>
            """,
            self.locale,
        )
        self.assertCodeEqual(
            code,
            """
            numberAttributeHtml : Locale.Locale -> { a | arg : Fluent.FluentNumber number } -> List (String, List (Html.Attribute msg)) -> List (Html.Html msg)
            numberAttributeHtml locale_ args_ attrs_ =
                [ Html.b (List.concat [ [ Attributes.attribute "data-foo" (Fluent.formatNumber locale_ args_.arg)
                                        ]
                                      , Fluent.selectAttributes attrs_ [ "b"
                                                                       , "[data-foo]"
                                                                       , "b[data-foo]"
                                                                       ]
                                      ]) [ Html.text "text"
                                         ]
                ]
            """,
        )
        self.assertEqual(errs, [])

    def test_argument(self):
        code, errs = compile_messages_to_elm(
            """
            hello-html = Hello <b>{ $username }</b>!
            """,
            self.locale,
            dynamic_html_attributes=False,
        )
        self.assertCodeEqual(
            code,
            """
            helloHtml : Locale.Locale -> { a | username : String } -> List (String, List (Html.Attribute msg)) -> List (Html.Html msg)
            helloHtml locale_ args_ attrs_ =
                [ Html.text "Hello "
                , Html.b [] [ Html.text args_.username
                            ]
                , Html.text "!"
                ]
             """,
        )
        self.assertEqual(errs, [])

    def test_non_string_args_in_text(self):
        code, errs = compile_messages_to_elm(
            """
            foo-html = <b>Text { $arg } { DATETIME($arg) }</b>
            """,
            self.locale,
            dynamic_html_attributes=False,
        )
        self.assertCodeEqual(
            code,
            """
            fooHtml : Locale.Locale -> { a | arg : Fluent.FluentDate } -> List (String, List (Html.Attribute msg)) -> List (Html.Html msg)
            fooHtml locale_ args_ attrs_ =
                [ Html.b [] [ Html.text (String.concat [ "Text "
                                                       , Fluent.formatDate locale_ args_.arg
                                                       , " "
                                                       , Fluent.formatDate locale_ args_.arg
                                                       ])
                            ]
                ]
             """,
        )
        self.assertEqual(errs, [])

    def test_text_message_call(self):
        code, errs = compile_messages_to_elm(
            """
            welcome-back = Welcome back!
            hello-html = Hello, friend! <b>{ welcome-back }</b>
            """,
            self.locale,
            dynamic_html_attributes=False,
        )
        self.assertCodeEqual(
            code,
            """
            welcomeBack : Locale.Locale -> a -> String
            welcomeBack locale_ args_ =
                "Welcome back!"

            helloHtml : Locale.Locale -> a -> List (String, List (Html.Attribute msg)) -> List (Html.Html msg)
            helloHtml locale_ args_ attrs_ =
                [ Html.text "Hello, friend! "
                , Html.b [] [ Html.text (welcomeBack locale_ args_)
                            ]
                ]
             """,
        )
        self.assertEqual(errs, [])

    def test_html_message_call(self):
        code, errs = compile_messages_to_elm(
            """
            welcome-html = Welcome to <b>Awesome site!</b>
            hello-html = Hello! { welcome-html }
            """,
            self.locale,
            dynamic_html_attributes=False,
        )
        self.assertCodeEqual(
            code,
            """
            welcomeHtml : Locale.Locale -> a -> List (String, List (Html.Attribute msg)) -> List (Html.Html msg)
            welcomeHtml locale_ args_ attrs_ =
                [ Html.text "Welcome to "
                , Html.b [] [ Html.text "Awesome site!"
                            ]
                ]

            helloHtml : Locale.Locale -> a -> List (String, List (Html.Attribute msg)) -> List (Html.Html msg)
            helloHtml locale_ args_ attrs_ =
                List.concat [ [ Html.text "Hello! "
                              ]
                            , welcomeHtml locale_ args_ attrs_
                            ]
             """,
        )
        self.assertEqual(errs, [])

    def test_text_message_call_attribute(self):
        code, errs = compile_messages_to_elm(
            """
            hello-html = Hello <b data-foo="&lt;stuff&gt; { hello-html.foo }">friend</b>
                      .foo = <xxx>
            """,
            self.locale,
            dynamic_html_attributes=False,
        )
        self.assertCodeEqual(
            code,
            """
            helloHtml : Locale.Locale -> a -> List (String, List (Html.Attribute msg)) -> List (Html.Html msg)
            helloHtml locale_ args_ attrs_ =
                [ Html.text "Hello "
                , Html.b [ Attributes.attribute "data-foo" (String.concat [ "<stuff> "
                                                                          , helloHtml_foo locale_ args_
                                                                          ])
                         ] [ Html.text "friend"
                           ]
                ]

            helloHtml_foo : Locale.Locale -> a -> String
            helloHtml_foo locale_ args_ =
                "<xxx>"
            """,
        )
        self.assertEqual(errs, [])

    def test_html_message_call_attribute(self):
        code, errs = compile_messages_to_elm(
            """
            hello-html = Hello <b class="{ foo-html }">friend</b>
            foo-html = Foo
            """,
            self.locale,
        )
        self.assertEqual(len(errs), 1)
        self.assertEqual(
            errs[0].args[0], "Cannot use HTML message foo-html from plain text context."
        )

    def test_html_message_call_from_plain_test(self):
        code, errs = compile_messages_to_elm(
            """
            hello = Hello { foo-html }
            foo-html = Foo
            """,
            self.locale,
        )
        self.assertEqual(len(errs), 1)
        self.assertEqual(
            errs[0].args[0], "Cannot use HTML message foo-html from plain text context."
        )

    def test_select_expression_1(self):
        # Test we get HTML handling of the pattern inside the select express
        # i.e HTML context propagates
        code, errs = compile_messages_to_elm(
            """
            hello-html = Hello { $gender ->
               [male]    <b>Mr. { $surname }</b>, nice to see you
               [female]  <b>Ms. { $surname }</b>, nice to see you
              *[other]   <b>{ $surname }</b>, nice to see you
             }
            """,
            self.locale,
            dynamic_html_attributes=False,
        )
        self.assertCodeEqual(
            code,
            """
            helloHtml : Locale.Locale -> { a | gender : String, surname : String } -> List (String, List (Html.Attribute msg)) -> List (Html.Html msg)
            helloHtml locale_ args_ attrs_ =
                List.concat [ [ Html.text "Hello "
                              ]
                            , case args_.gender of
                                  "male" ->
                                      [ Html.b [] [ Html.text (String.concat [ "Mr. "
                                                                             , args_.surname
                                                                             ])
                                                  ]
                                      , Html.text ", nice to see you"
                                      ]
                                  "female" ->
                                      [ Html.b [] [ Html.text (String.concat [ "Ms. "
                                                                             , args_.surname
                                                                             ])
                                                  ]
                                      , Html.text ", nice to see you"
                                      ]
                                  _ ->
                                      [ Html.b [] [ Html.text args_.surname
                                                  ]
                                      , Html.text ", nice to see you"
                                      ]

                            ]
             """,
        )
        self.assertEqual(errs, [])

    def test_html_term_inline(self):
        code, errs = compile_messages_to_elm(
            """
            welcome-html = Welcome to { -brand-html }
            -brand-html = Awesomeness<sup>2</sup>
            """,
            self.locale,
            dynamic_html_attributes=False,
        )
        self.assertCodeEqual(
            code,
            """
            welcomeHtml : Locale.Locale -> a -> List (String, List (Html.Attribute msg)) -> List (Html.Html msg)
            welcomeHtml locale_ args_ attrs_ =
                [ Html.text "Welcome to Awesomeness"
                , Html.sup [] [ Html.text "2"
                              ]
                ]
            """,
        )
        self.assertEqual(errs, [])

    def test_plain_term_inline(self):
        code, errs = compile_messages_to_elm(
            """
            welcome-html = Welcome to <b>{ -brand }</b>
            -brand = Awesomeness2
            """,
            self.locale,
            dynamic_html_attributes=False,
        )
        self.assertCodeEqual(
            code,
            """
            welcomeHtml : Locale.Locale -> a -> List (String, List (Html.Attribute msg)) -> List (Html.Html msg)
            welcomeHtml locale_ args_ attrs_ =
                [ Html.text "Welcome to "
                , Html.b [] [ Html.text "Awesomeness2"
                            ]
                ]
            """,
        )
        self.assertEqual(errs, [])



"""
Utilities for doing Python code generation
"""
from __future__ import absolute_import, unicode_literals

import contextlib
import re

import six

from elm_fluent import types

# This module provides simple utilities for building up Elm source code. It
# implements only what is really needed by compiler.py, with a number of aims
# and constraints:
#
# 1. Performance.
#
#    The resulting Elm code should do as little as possible, especially for
#    simple cases (which are by far the most common for .ftl files)
#
# 2. Correctness (obviously)
#
#    In particular, we should try to make it hard to generate incorrect code,
#    esepcially incorrect code that is syntactically correct and therefore
#    compiles but doesn't work. In particular, we try to make it hard to
#    generate accidental name clashes, or use variables that are not defined.
#
# 3. Simplicity
#
#    The resulting Elm code should be easy to read and understand.
#
# 4. Predictability
#
#    Since we want to test the resulting source code, we have made some design
#    decisions that aim to ensure things like function argument names are
#    consistent and so can predicted easily.


class ElmAst(object):
    def __init__(self, from_ftl_source=None):
        self.from_ftl_source = from_ftl_source

    def simplify(self, changes):
        """
        Simplify the statement/expression, returning either a modified
        self, or a new object.

        This method should call .simplify(changes) on any contained subexpressions
        or statements.

        If changes were made, a True value must be appended to the passed in changes list.
        """
        return self

    def sub_expressions(self):
        """
        Returns an iterable of all syntax nodes contained in this node
        """
        raise NotImplementedError(
            "{0} needs to implement sub_expressions".format(self.__class__)
        )

    def finalize(self):
        pass

    def as_source_code(self):
        builder = SourceCodeBuilder()
        self.build_source(builder)
        return builder.render()


class Variables(object):
    def __init__(self, scope):
        self.scope = scope

    def __getitem__(self, item):
        return VariableReference(item, self.scope)


class Scope(ElmAst):
    is_default_imports = False

    def __init__(self, parent_scope=None):
        super(Scope, self).__init__()
        self.parent_scope = parent_scope
        self.names = set()
        self._function_arg_reserved_names = set()
        self._types = {}
        self.variables = Variables(self)

    def names_in_use(self):
        names = self.names
        if self.parent_scope is not None:
            names = names | self.parent_scope.names_in_use()
        return names

    def function_arg_reserved_names(self):
        names = self._function_arg_reserved_names
        if self.parent_scope is not None:
            names = names | self.parent_scope.function_arg_reserved_names()
        return names

    def all_reserved_names(self):
        return self.names_in_use() | self.function_arg_reserved_names()

    def reserve_name(self, requested, function_arg=False, type=None):
        """
        Reserve a name as being in use in a scope.

        Pass function_arg=True if this is a function argument.
        'properties' is an optional dict of additional properties
        (e.g. the type associated with a name)
        """

        def _add(final):
            self.names.add(final)
            if type is not None:
                self._types[final] = type
            return final

        if function_arg:
            if requested in self.function_arg_reserved_names():
                assert requested not in self.names_in_use()
                return _add(requested)
            else:
                if requested in self.all_reserved_names():
                    raise AssertionError(
                        "Cannot use '{0}' as argument name as it is already in use".format(
                            requested
                        )
                    )

        cleaned = cleanup_name(requested)

        attempt = cleaned
        count = 2  # instance without suffix is regarded as 1
        # To avoid shadowing of global names in local scope, we
        # take into account parent scope when assigning names.

        used = self.all_reserved_names()
        while attempt in used:
            attempt = cleaned + str(count)
            count += 1
        return _add(attempt)

    def reserve_function_arg_name(self, name):
        """
        Reserve a name for *later* use as a function argument. This does not result
        in that name being considered 'in use' in the current scope, but will
        avoid the name being assigned for any use other than as a function argument.
        """
        # To keep things simple, and the generated code predictable, we reserve
        # names for all function arguments in a separate scope, and insist on
        # the exact names
        if name in self.all_reserved_names():
            raise AssertionError(
                "Can't reserve '{0}' as function arg name as it is already reserved".format(
                    name
                )
            )
        self._function_arg_reserved_names.add(name)

    def get_type(self, name):
        """
        Gets a dictionary of properties for the name.
        Raises exception if the name is not reserved in this scope or parent
        """
        if name in self._types:
            return self._types[name]
        else:
            if self.parent_scope is None:
                raise KeyError(name)
            return self.parent_scope.get_type(name)

    def set_type(self, name, type):
        """
        Sets the type for the name.
        Raises exception if the name is not reserved in this scope or parent.
        """
        scope = self
        while True:
            if name in scope._types:
                scope._types[name] = type
                break
            else:
                scope = scope.parent_scope

    def get_imported_module(self, name):
        # Chain back until we find a 'Module', which has an implementation
        # for this.
        return self.parent_scope.get_imported_module(name)

    def get_name_qualifier(self, module):
        """
        For a given module, find the local name of that module in this scope
        """
        # Chain back until we a find a 'Module', which has an implementation for this
        return self.parent_scope.get_name_qualifier(module)


_IDENTIFIER_SANITIZER_RE = re.compile("[^a-zA-Z0-9_]")
_IDENTIFIER_START_RE = re.compile("^[a-zA-Z]")


def cleanup_name(name):
    # Choose safe subset of known allowed chars
    name = _IDENTIFIER_SANITIZER_RE.sub("", name)
    if not _IDENTIFIER_START_RE.match(name):
        name = "n" + name
    return name


# Keywords: see
# https://github.com/elm/compiler/blob/master/compiler/src/Parse/Primitives/Keyword.hs#L57
# However, some seem to be accepted fine as function names, they are only
# keywords in some positions it seems. These are commented out.
ELM_KEYWORDS = set(
    [
        "type",
        # "alias",
        "port",
        "if",
        "then",
        "else",
        "case",
        "of",
        "let",
        "in",
        "infix",
        # "left",
        # "right",
        # "non",
        "module",
        "import",
        "exposing",
        "as",
        "where",
        # "effect",
        # "command",
        # "subscription",
        # "true",
        # "false",
        # "null",
    ]
)


class Module(Scope):
    def __init__(self, name=None):
        from .stubs.defaults import default_imports

        super(Module, self).__init__(parent_scope=default_imports)
        self.statements = {}  # Dict from statement number to statement
        self.exports = []
        self.import_dict = {}  # Map from local name to module
        self.name = name

    def __repr__(self):
        return "<Module {0}>".format(self.name)

    def all_reserved_names(self):
        return super(Module, self).all_reserved_names() | ELM_KEYWORDS

    def get_imported_module(self, import_name):
        # Here to avoid circular imports. TODO CLEANUP - more generic mechanism for
        # default imports, while still avoid circular import problem.
        if import_name == "String":
            from .stubs.string import module as string_module

            return string_module
        return self.import_dict[import_name]

    def get_name_qualifier(self, module):
        """
        For a given module, find the qualifer that must be used for that module in this module
        """
        if module is self:
            return ""

        if module.is_default_imports:
            # defaults need no qualifiers
            return ""
        for name, mod in self.import_dict.items():
            if mod == module:
                return name + "."
        raise LookupError(
            "module {0} not found in {1}. Missing 'add_import'?".format(module, self)
        )

    def add_function(self, func_name, func, expose=True, source_order=None):
        assert func.func_name == func_name
        self.exports.append(func_name)

        if source_order is None:
            if not self.statements:
                source_order = 0
            else:
                source_order = max(self.statements.keys()) + 1
        self.statements[source_order] = func

    def add_import(self, module, name):
        self.import_dict[name] = module

    def as_source_code(self, include_module_line=True, include_imports=True):
        lines = []
        if include_module_line:
            assert self.name is not None
            lines.append(
                "module {0} exposing ({1})\n".format(self.name, ", ".join(self.exports))
            )
            lines.append("\n")
        if include_imports and self.import_dict:
            for name, module in sorted(
                self.import_dict.items(), key=lambda pair: pair[1].name
            ):
                if self.import_is_used(name, module):
                    # We only support 'as' imports, to avoid name conflicts
                    # with functions defined in the module
                    lines.append("import {0} as {1}\n".format(module.name, name))
            lines.append("\n")

        for s in self.sorted_statements():
            lines.append(s.as_source_code())
            # blank line
            lines.append("\n\n")
        return "".join(lines)

    def sorted_statements(self):
        return [s for i, s in sorted(self.statements.items(), key=lambda pair: pair[0])]

    def import_is_used(self, local_name, module):
        for node in traverse(self):
            # Check for variable references
            if isinstance(node, VariableReference):
                if node.module_name == local_name:
                    return True

            # Check function signatures
            if isinstance(node, Function):
                function_type = node.type
                if function_type is not None:
                    for t in types.signature_traverse(function_type):
                        if getattr(t, "module", None) == module:
                            return True

        return False

    def simplify(self, changes):
        self.statements = {i: s.simplify(changes) for i, s in self.statements.items()}
        return self

    def sub_expressions(self):
        return self.sorted_statements()


class Statement(ElmAst):
    pass


class _Assignment(Statement):
    def __init__(self, name, value):
        self.name = name
        self.value = value

    # def format_names(self):
    #     if len(self.names) == 1:
    #         return self.names[0]
    #     else:
    #         return "({0})".format(", ".join(n for n in self.names))

    def build_source(self, builder):
        builder.add_part(self.name)
        builder.add_part(" = ")
        self.value.build_source(builder)

    def simplify(self, changes):
        self.value = self.value.simplify(changes)
        return self

    def sub_expressions(self):
        yield self.value


class Function(Scope, Statement):
    def __init__(self, name, args=None, parent_scope=None):
        super(Function, self).__init__(parent_scope=parent_scope)
        self.func_name = name
        self.body = Let(parent_scope=self)
        if args is None:
            args = ()
        self.args = args

        # Get the types of the arguments from the registered function type
        remaining_function_type = self.type
        for arg in args:
            if arg in parent_scope.names_in_use():
                raise AssertionError(
                    "Can't use '{0}' as function argument name because it shadows other names".format(
                        arg
                    )
                )
            if remaining_function_type is not None:
                arg_type = remaining_function_type.input_type
                remaining_function_type = remaining_function_type.output_type
            else:
                arg_type = None
            self.reserve_name(arg, function_arg=True, type=arg_type)

    @property
    def type(self):
        try:
            return self.parent_scope.get_type(self.func_name)
        except KeyError:
            return None

    @type.setter
    def type(self, type_obj):
        self.parent_scope.set_type(self.func_name, type_obj.constrain(self.type))

    def sub_expressions(self):
        yield self.body

    def finalize(self):
        t = self.type
        if t is not None:
            args = self.args
            output_type = t
            for a in args:
                output_type = output_type.output_type
            self.body.constrain_type(output_type)

    def build_source(self, builder):
        if self.parent_scope is None:
            signature = ""
        else:
            function_type = self.type
            if function_type is None:
                signature = ""
            else:
                signature = "{name} : {signature}\n".format(
                    signature=function_type.as_signature(self.parent_scope),
                    name=self.func_name,
                )

        builder.add_part(signature)
        builder.add_part(self.func_name)
        for arg in self.args:
            builder.add_part(" " + arg)

        builder.add_part(" =\n")
        with builder.indented():
            self.body.build_source(builder)

    def simplify(self, changes):
        self.body = self.body.simplify(changes)
        return self


def resolve_type(type_object_or_name):
    # To avoid circular imports, we accept a string to represent types
    # in dtypes
    from .stubs import defaults as dtypes

    if isinstance(type_object_or_name, six.text_type):
        return getattr(dtypes, type_object_or_name)
    else:
        return type_object_or_name


def fixed_type(type_object_or_name):
    class FixedType(object):
        _type_object_or_name = type_object_or_name

        @property
        def type(self):
            return resolve_type(self._type_object_or_name)

    return FixedType


class SourceCodeBuilder(object):
    BLOCK_INDENT_SIZE = 4

    def __init__(self):
        self.parts = []
        self.indent_level = 0
        self.current_line = ""

    def add_part(self, part):
        if self.current_line == "":
            # Add indentation:
            indent = " " * self.indent_level
            self.parts.append(indent)
            self.current_line += indent
        if "\n" in part and part.find("\n") != len(part) - 1:
            raise ValueError(
                "If you pass '\n' to add_part, it must be at the end of string value. Received {0}".format(
                    repr(part)
                )
            )
        self.current_line += part
        if self.current_line.endswith("\n"):
            self.current_line = ""
        self.parts.append(part)

    def render(self):
        return "".join(self.parts)

    @contextlib.contextmanager
    def aligned_block(self):
        old_indent_level = self.indent_level
        if self.current_line == "":
            alignment_level = self.indent_level
        else:
            alignment_level = len(self.current_line)
        self.indent_level = alignment_level
        yield
        self.indent_level = old_indent_level

    @contextlib.contextmanager
    def indented(self):
        self.indent_level += self.BLOCK_INDENT_SIZE
        yield self
        self.indent_level -= self.BLOCK_INDENT_SIZE

    @contextlib.contextmanager
    def parens_if_needed(self, expr):
        # TODO - may need a better way to determine when we need parentheses, that
        # respects operator precedence etc. At the moment the following suffices.
        # Operators do wrapping always to be safe.
        wrapping = not isinstance(
            expr, (Literal, Bracketing, VariableReference, AttributeReference)
        )
        if wrapping:
            self.add_part("(")

        yield self

        if wrapping:
            self.add_part(")")


class Expression(ElmAst):
    # type represents the Elm type this expression will produce,
    # if we know it (UnconstrainedType otherwise).
    @property
    def type(self):
        return types.UnconstrainedType()

    def constrain_type(self, type_obj, from_ftl_source=None):
        raise NotImplementedError(
            "Object {0} of type {1} does not implement constrain_type".format(
                self, type(self)
            )
        )

    # Python 2.7 compatible kwarg syntax
    def apply(self, *args, **kwargs):
        """
        Function application
        """
        from_ftl_source = kwargs.pop("from_ftl_source", None)
        assert not kwargs, "Unexpected keyword args {0}".format(
            ", ".join(kwargs.keys())
        )
        return FunctionCall(self, args, from_ftl_source=from_ftl_source)


class Let(Expression, Scope):
    def __init__(self, parent_scope=None):
        Scope.__init__(self, parent_scope=parent_scope)
        self.value = None
        self.assignments = []

    @property
    def type(self):
        assert self.value is not None
        return self.value.type

    def constrain_type(self, type_obj, from_ftl_source=None):
        return self.value.constrain_type(type_obj.constrain(self.value.type))

    def add_assignment(self, name, value):
        """
        Adds an assigment of the form:

           x = value

        """
        # If needed, we might have to add assignment from tuples here
        assigned_name = self.reserve_name(name, type=value.type)
        self.assignments.append(_Assignment(assigned_name, value))
        return self.variables[name]

    def simplify(self, changes):
        self.value = self.value.simplify(changes)
        for assignment in self.assignments:
            assignment.simplify(changes)
        if len(self.assignments) == 0:
            changes.append(True)
            return self.value
        if len(self.assignments) == 1:
            if isinstance(self.value, VariableReference):
                if self.value.name == self.assignments[0].name:
                    changes.append(True)
                    return self.assignments[0].value
        return self

    def sub_expressions(self):
        return self.assignments + [self.value]

    def build_source(self, builder):
        with builder.aligned_block():
            builder.add_part("let\n")
            with builder.indented():
                for a in self.assignments:
                    a.build_source(builder)
                    builder.add_part("\n")
            builder.add_part("in\n")
            with builder.indented():
                self.value.build_source(builder)


class If(Expression):
    def __init__(
        self, condition=None, true_branch=None, false_branch=None, parent_scope=None
    ):
        self.condition = condition
        self.true_branch = true_branch or Let(parent_scope=parent_scope)
        self.false_branch = false_branch or Let(parent_scope=parent_scope)

    @property
    def type(self):
        # TODO - can we do better than just returning the true branch
        # Does it matter?
        return self.true_branch.type

    def constrain_type(self, type_obj, from_ftl_source=None):
        self.true_branch.constrain_type(
            type_obj.constrain(self.true_branch.type), from_ftl_source=from_ftl_source
        )
        self.false_branch.constrain_type(
            type_obj.constrain(self.false_branch.type), from_ftl_source=from_ftl_source
        )

    def build_source(self, builder):
        with builder.aligned_block():
            builder.add_part("if ")
            self.condition.build_source(builder)
            builder.add_part(" then\n")
            with builder.indented():
                self.true_branch.build_source(builder)
                builder.add_part("\n")
            builder.add_part("else\n")
            with builder.indented():
                self.false_branch.build_source(builder)

    def sub_expressions(self):
        yield self.condition
        yield self.true_branch
        yield self.false_branch

    def simplify(self, changes):
        self.condition = self.condition.simplify(changes)
        self.true_branch = self.true_branch.simplify(changes)
        self.false_branch = self.false_branch.simplify(changes)
        return self


class Case(Expression):
    def __init__(self, selector=None, parent_scope=None):
        self.parent_scope = parent_scope
        self.selector = selector
        self.branches = []

    def add_branch(self, matcher):
        value = Let(self.parent_scope)
        self.branches.append((matcher, value))
        return value

    def build_source(self, builder):
        with builder.aligned_block():
            builder.add_part("case ")
            self.selector.build_source(builder)
            builder.add_part(" of\n")
            with builder.indented():
                for matcher, value in self.branches:
                    matcher.build_source(builder)
                    builder.add_part(" ->\n")
                    with builder.indented():
                        value.build_source(builder)
                        builder.add_part("\n")

    @property
    def type(self):
        assert len(self.branches) > 0
        # TODO - can we do better than just returning the first one?
        # Does it matter?
        return self.branches[0][1].type

    def constrain_type(self, type_obj, from_ftl_source=None):
        for m, val in self.branches:
            val.constrain_type(
                type_obj.constrain(val.type), from_ftl_source=from_ftl_source
            )

    def sub_expressions(self):
        yield self.selector
        for matcher, value in self.branches:
            yield matcher
            yield value

    def simplify(self, changes):
        self.selector = self.selector.simplify(changes)
        self.branches = [
            (matcher.simplify(changes), value.simplify(changes))
            for matcher, value in self.branches
        ]
        return self


class Literal(Expression):
    def sub_expressions(self):
        return []

    def constrain_type(self, type_obj, from_ftl_source=None):
        t = type_obj.constrain(self.type)
        assert t is self.type, "Expected {0} is {1}".format(t, self.type)


class Bracketing(object):
    """
    Sentinel class to indicate expression that do their own bracketing
    so don't need extra parenthesis
    """

    pass


class String(fixed_type("String"), Literal):
    def __init__(self, string_value):
        self.string_value = string_value

    def __repr__(self):
        return "<String {0}>".format(repr(self.string_value))

    def build_source(self, builder):
        # TODO - escapes for other chars?
        builder.add_part(
            '"{0}"'.format(self.string_value.replace('"', '\\"').replace("\n", "\\n"))
        )


class Number(fixed_type("Number"), Literal):
    def __init__(self, number):
        self.number = number

    def build_source(self, builder):
        # TODO - Are there cases where this won't work?
        builder.add_part(repr(self.number))


class List(Bracketing, Expression):
    def __init__(self, items):
        self.items = items

    @property
    def type(self):
        # TODO - can we do better than just using the first one?
        # Does it matter?
        from .stubs import defaults as dtypes

        if self.items:
            return dtypes.List.specialize(a=self.items[0].type)
        else:
            return dtypes.List

    def constrain_type(self, type_obj, from_ftl_source=None):
        list_type = type_obj.constrain(self.type)
        val_type = list_type.param_dict["a"]
        for val in self.items:
            val.constrain_type(
                val_type.constrain(val.type), from_ftl_source=from_ftl_source
            )

    def build_source(self, builder):
        if len(self.items) == 0:
            builder.add_part("[]")
        else:
            with builder.aligned_block():
                for i, item in enumerate(self.items):
                    if i == 0:
                        builder.add_part("[ ")
                    else:
                        builder.add_part(", ")
                    item.build_source(builder)
                    builder.add_part("\n")
                builder.add_part("]")

    def sub_expressions(self):
        return self.items

    def simplify(self, changes):
        self.items = [item.simplify(changes) for item in self.items]
        return self


class Concat(Expression):
    def __init__(self, parts, from_ftl_source=None):
        super(Concat, self).__init__(from_ftl_source=from_ftl_source)
        self.parts = parts

    def __repr__(self):
        return "<Concat {0}>".format(repr(self.parts))

    def sub_expressions(self):
        return self.parts

    def constrain_type(self, type_obj, from_ftl_source=None):
        for part in self.parts:
            part.constrain_type(
                type_obj.constrain(part.type), from_ftl_source=from_ftl_source
            )

    def build_source(self, builder):
        builder.add_part(self.function_call + " ")
        List(self.parts).build_source(builder)

    def simplify(self, changes):
        # Simplify sub parts (while eliminating empty)
        self.parts = [
            part.simplify(changes) for part in self.parts if not self.is_empty(part)
        ]

        # Merge adjacent List(like) objects.
        new_parts = []
        for part in self.parts:
            if (
                len(new_parts) > 0
                and isinstance(new_parts[-1], self.literal)
                and isinstance(part, self.literal)
            ):
                new_parts[-1] = self.merge_two(new_parts[-1], part)
            else:
                new_parts.append(part)
        if len(new_parts) < len(self.parts):
            changes.append(True)
        self.parts = new_parts

        # See if we can eliminate the Concat altogether
        if len(self.parts) == 0:
            changes.append(True)
            return self.empty()
        elif len(self.parts) == 1:
            changes.append(True)
            return self.parts[0]
        else:
            return self


class StringConcat(fixed_type("String"), Concat):
    literal = String
    function_call = "String.concat"

    def is_empty(self, part):
        return isinstance(part, String) and part.string_value == ""

    def empty(self):
        return self.literal("")

    def merge_two(self, part1, part2):
        return self.literal(part1.string_value + part2.string_value)


class ListConcat(Concat):
    def __init__(self, parts, type_obj):
        self.parts = parts
        self._type = type_obj

    @property
    def type(self):
        return self._type

    def constrain_type(self, type_obj, from_ftl_source=None):
        assert type_obj == self._type, "Expected {0} == {1}".format(
            type_obj, self._type
        )
        for p in self.parts:
            p.constrain_type(type_obj, from_ftl_source=from_ftl_source)

    literal = List
    function_call = "List.concat"

    def is_empty(self, part):
        return isinstance(part, List) and len(part.items) == 0

    def empty(self):
        return self.literal([])

    def merge_two(self, part1, part2):
        return self.literal(part1.items + part2.items)


class VariableReference(Expression):
    # Instead of using this directly, normally use
    # scope.variables[qualified_name] for convenience and readability
    def __init__(self, qualified_name, scope):
        super(VariableReference, self).__init__()
        if "." in qualified_name:
            module_name, name = qualified_name.split(".")
        else:
            module_name = None
            name = qualified_name
        if module_name is not None:
            definition_scope = scope.get_imported_module(module_name)
        else:
            definition_scope = scope
        if name not in definition_scope.names_in_use():
            raise AssertionError("Cannot refer to undefined name '{0}'".format(name))

        self._definition_scope = definition_scope
        self.module_name = module_name
        self.name = name

    @property
    def type(self):
        return self._definition_scope.get_type(self.name)

    def constrain_type(self, type_obj, from_ftl_source=None):
        self._definition_scope.set_type(self.name, type_obj.constrain(self.type))

    def build_source(self, builder):
        builder.add_part(
            ((self.module_name + ".") if self.module_name is not None else "")
            + self.name
        )

    def sub_expressions(self):
        return []

    def __repr__(self):
        return "<VariableReference {0}>".format(self.as_source_code())


class AttributeReference(Expression):
    def __init__(self, variable, attribute_name):
        self.variable = variable
        self.attribute_name = attribute_name
        assert isinstance(variable.type, types.Record)
        variable.type.add_field(attribute_name)

    @property
    def type(self):
        return self.variable.type.fields[self.attribute_name]

    def constrain_type(self, type_obj, from_ftl_source=None):
        self.variable.type.add_field(
            self.attribute_name, type_obj, from_ftl_source=from_ftl_source
        )

    def __repr__(self):
        return "<AttributeReference {0}.{1}>".format(
            repr(self.variable), self.attribute_name
        )

    def build_source(self, builder):
        self.variable.build_source(builder)
        builder.add_part(".")
        builder.add_part(self.attribute_name)

    def sub_expressions(self):
        yield self.variable


class FunctionCall(Expression):
    # Instead of using this directly, normally use expr.apply()
    def __init__(self, expr, args, from_ftl_source=None):
        super(FunctionCall, self).__init__(from_ftl_source=from_ftl_source)
        self.expr = expr
        self.args = args
        self._type = self.expr.type.apply_args(
            self.args, from_ftl_source=from_ftl_source
        )

    @property
    def type(self):
        return self._type

    def constrain_type(self, type_obj, from_ftl_source=None):
        # Currently all our functions have fully concrete return types,
        # so we don't need anything beyond this:
        assert self.type == type_obj, "Expected {0} == {1}".format(self.type, type_obj)

    def __repr__(self):
        return "<FunctionCall {0} {1}>".format(
            self.expr.as_source_code(), " ".join(repr(a) for a in self.args)
        )

    def build_source(self, builder):
        self.expr.build_source(builder)
        for arg in self.args:
            builder.add_part(" ")
            with builder.parens_if_needed(arg):
                arg.build_source(builder)

    def sub_expressions(self):
        return [self.expr] + list(self.args)

    def simplify(self, changes):
        self.args = [arg.simplify(changes) for arg in self.args]
        return self


class Otherwise(Expression):
    def __init__(self):
        self._type = types.UnconstrainedType()

    @property
    def type(self):
        return self._type

    def constrain_type(self, type_obj, from_ftl_source=None):
        self._type = type_obj.constrain(self._type)

    def sub_expressions(self):
        return []

    def build_source(self, builder):
        builder.add_part("_")


class CompilationError(Expression):
    def __init__(self, type_obj=None):
        if type_obj is None:
            type_obj = types.UnconstrainedType()
        self._type = type_obj

    @property
    def type(self):
        return self._type

    def constrain_type(self, type_obj, from_ftl_source=None):
        # Just assign without complaining
        self._type = type_obj

    def build_source(self, builder):
        # Return something that will definitely cause the Elm code to fail to
        # compile. Normally the value is not written out anywhere. It exists as
        # an Expression to allow compilation to continue, so that as many error
        # messages as possible can be collected instead of quitting early.
        builder.add_part("!!!COMPILATION_ERROR!!!")

    def sub_expressions(self):
        return []


def infix_operator(operator, return_type, operand_type):
    class Op(fixed_type(return_type), Bracketing, Expression):
        def __init__(self, left, right):
            self.left = left
            self.right = right

        def build_source(self, builder):
            builder.add_part("(")
            self.left.build_source(builder)
            builder.add_part(" " + operator + " ")
            self.right.build_source(builder)
            builder.add_part(")")

        def simplify(self, changes):
            self.left = self.left.simplify(changes)
            self.right = self.right.simplify(changes)
            return self

        def sub_expressions(self):
            yield self.left
            yield self.right

        def constrain_type(self, type_obj, from_ftl_source=None):
            self.left.constrain_type(type_obj.constrain(resolve_type(operand_type)))
            self.right.constrain_type(type_obj.constrain(resolve_type(operand_type)))

    return Op


Equals = infix_operator("==", "Bool", types.UnconstrainedType())
Add = infix_operator("+", "Number", "Number")


class RecordUpdate(Bracketing, Expression):
    def __init__(self, var, **updates):
        assert isinstance(
            var.type, types.Record
        ), "isinstance({0}, types.Record)".format(var.type)
        assert isinstance(
            var, VariableReference
        ), "isinstance({0}, VariableReference)".format(var)
        assert (
            var.module_name is None
        ), "Record update syntax does not allow qualified name like {0}.{1}".format(
            var.module_name, var.name
        )
        self.var = var
        self.updates = updates
        for name, val in updates.items():
            var.type.add_field(name, val.type)

    @property
    def type(self):
        return self.var.type

    def constrain_type(self, type_obj, from_ftl_source=None):
        self.var.constrain_type(
            type_obj.constrain(self.var.type), from_ftl_source=from_ftl_source
        )

    def sub_expressions(self):
        for name, val in self.updates.items():
            yield val

    def build_source(self, builder):
        builder.add_part("{ ")
        self.var.build_source(builder)
        builder.add_part(" | ")
        for i, (name, val) in enumerate(sorted(self.updates.items())):
            builder.add_part(name)
            builder.add_part(" = ")
            val.build_source(builder)
            if i < len(self.updates) - 1:
                builder.add_part(", ")
        builder.add_part(" }")


def simplify(source_code):
    # Can't use 'traverse' for simplify because of the way it needs to change
    # the actual tree itself.
    finalize(source_code)

    changes = [True]
    while any(changes):
        changes = []
        source_code = source_code.simplify(changes)
    return source_code


def traverse(node):
    sub_parts = node.sub_expressions()
    for part in sub_parts:
        for t in traverse(part):
            yield t
    yield node


def finalize(source_code):
    for node in traverse(source_code):
        node.finalize()



"""
Stub to define types for the Intl.Locale module
"""
from __future__ import absolute_import, unicode_literals

from .. import codegen, types
from . import defaults as dtypes

module = codegen.Module(name="Intl.Locale")

Locale = types.Type("Locale", module)

module.reserve_name("toLanguageTag", type=types.Function(Locale, dtypes.String))



# -*- coding: utf-8 -*-

from __future__ import absolute_import, unicode_literals

import textwrap
import unittest

from elm_fluent import codegen, types
from elm_fluent.stubs import defaults as dtypes


def normalize_elm(text):
    return textwrap.dedent(text.rstrip()).strip()


class TestCodeGen(unittest.TestCase):
    def assertCodeEqual(self, code1, code2):
        self.assertEqual(normalize_elm(code2), normalize_elm(code1))

    def test_module_builtins(self):
        module = codegen.Module()
        self.assertIn("min", module.all_reserved_names())

    def test_module_keywords(self):
        module = codegen.Module()
        self.assertIn("import", module.all_reserved_names())

    def test_reserve_name(self):
        scope = codegen.Scope()
        name1 = scope.reserve_name("name")
        name2 = scope.reserve_name("name")
        self.assertEqual(name1, "name")
        self.assertNotEqual(name1, name2)
        self.assertEqual(name2, "name2")

    def test_variables(self):
        scope = codegen.Scope()
        name = scope.reserve_name("name")
        var = scope.variables[name]
        self.assertEqual(type(var), codegen.VariableReference)
        self.assertEqual(var.name, name)

    def test_reserve_name_function_arg_disallowed(self):
        scope = codegen.Scope()
        scope.reserve_name("name")
        self.assertRaises(AssertionError, scope.reserve_name, "name", function_arg=True)

    def test_reserve_name_function_arg(self):
        scope = codegen.Scope()
        scope.reserve_function_arg_name("arg_name")
        scope.reserve_name("myfunc")
        func = codegen.Function("myfunc", args=["arg_name"], parent_scope=scope)
        self.assertNotIn("arg_name2", func.all_reserved_names())

    def test_reserve_name_nested(self):
        parent = codegen.Scope()
        parent_name = parent.reserve_name("name")
        self.assertEqual(parent_name, "name")

        child1 = codegen.Scope(parent_scope=parent)
        child2 = codegen.Scope(parent_scope=parent)

        child1_name = child1.reserve_name("name")
        self.assertNotEqual(child1_name, parent_name)

        child2_name = child2.reserve_name("name")
        self.assertNotEqual(child2_name, parent_name)

        # But children can have same names, they don't shadow each other.
        # To be deterministic, we expect the same name
        self.assertEqual(child1_name, child2_name)

    def test_reserve_name_after_reserve_function_arg(self):
        scope = codegen.Scope()
        scope.reserve_function_arg_name("my_arg")
        name = scope.reserve_name("my_arg")
        self.assertEqual(name, "my_arg2")

    def test_reserve_function_arg_after_reserve_name(self):
        scope = codegen.Scope()
        scope.reserve_name("my_arg")
        self.assertRaises(AssertionError, scope.reserve_function_arg_name, "my_arg")

    def test_get_type(self):
        scope = codegen.Scope()
        scope.reserve_name("name", type=dtypes.Bool)
        self.assertEqual(scope.get_type("name"), dtypes.Bool)

    def test_function(self):
        module = codegen.Module()
        func = codegen.Function(
            "myfunc", args=["myarg1", "myarg2"], parent_scope=module
        )
        func.body.value = codegen.String("hello")
        func = codegen.simplify(func)
        self.assertCodeEqual(
            func.as_source_code(),
            """
            myfunc myarg1 myarg2 =
                "hello"
        """,
        )

    def test_function_typed(self):
        module = codegen.Module()
        function_type = types.Function.for_multiple_inputs(
            [dtypes.String, dtypes.Number], dtypes.String
        )
        module.reserve_name("myfunc", type=function_type)
        func = codegen.Function(
            "myfunc", args=["myarg1", "myarg2"], parent_scope=module
        )
        func.body.value = codegen.String("hello")
        func = codegen.simplify(func)
        self.assertCodeEqual(
            func.as_source_code(),
            """
            myfunc : String -> number -> String
            myfunc myarg1 myarg2 =
                "hello"
        """,
        )
        self.assertEqual(func.variables["myarg2"].type, dtypes.Number)

    def test_attribute_reference_in_function(self):
        module = codegen.Module()
        record_type = types.Record()
        function_type = types.Function(record_type, dtypes.String)
        module.reserve_name("myfunc", type=function_type)
        func = codegen.Function("myfunc", args=["myarg"], parent_scope=module)
        func.body.value = codegen.AttributeReference(func.variables["myarg"], "myattr")
        func.finalize()
        codegen.simplify(func)
        self.assertCodeEqual(
            func.as_source_code(),
            """
            myfunc : { a | myattr : String } -> String
            myfunc myarg =
                myarg.myattr
        """,
        )

    def test_let_no_assignments(self):
        module = codegen.Module()
        func = codegen.Function("myfunc", parent_scope=module)
        func.body.value = codegen.String("Hello")
        func = codegen.simplify(func)
        self.assertCodeEqual(
            func.as_source_code(),
            """
            myfunc =
                "Hello"
        """,
        )

    def test_let_one_assignment(self):
        module = codegen.Module()
        func = codegen.Function("myfunc", parent_scope=module)
        let = func.body
        x = let.add_assignment("x", codegen.String("Hello"))
        self.assertEqual(x.name, "x")
        self.assertEqual(x.type, dtypes.String)
        let.value = x
        func = codegen.simplify(func)
        self.assertCodeEqual(
            func.as_source_code(),
            """
            myfunc =
                "Hello"
        """,
        )

    def test_let_two_assignments(self):
        module = codegen.Module()
        func = codegen.Function("myfunc", parent_scope=module)
        let = func.body
        x = let.add_assignment("x", codegen.Number(1))
        y = let.add_assignment("y", codegen.Number(2))
        let.value = codegen.Add(x, y)
        func = codegen.simplify(func)
        # TODO SOMEDAY - remove unnecessary parenthesis in final expression
        self.assertCodeEqual(
            func.as_source_code(),
            """
            myfunc =
                let
                    x = 1
                    y = 2
                in
                    (x + y)
            """,
        )

    def test_add_function(self):
        module = codegen.Module(name="Main")
        func_name = module.reserve_name("myfunc")
        func = codegen.Function(func_name, parent_scope=module)
        func.body.value = codegen.String("hi")
        module.add_function(func_name, func)
        func = codegen.simplify(func)
        self.assertCodeEqual(
            module.as_source_code(),
            """
            module Main exposing (myfunc)

            myfunc =
                "hi"
        """,
        )

    def test_variable_reference(self):
        module = codegen.Module()
        name = module.reserve_name("name")
        ref = codegen.VariableReference(name, module)
        self.assertEqual(ref.as_source_code(), "name")

    def test_variable_reference_check(self):
        module = codegen.Module()
        self.assertRaises(AssertionError, codegen.VariableReference, "name", module)

    def test_variable_reference_function_arg_check(self):
        module = codegen.Module()
        func_name = module.reserve_name("myfunc")
        func = codegen.Function(func_name, args=["my_arg"], parent_scope=module)
        # Can't use undefined 'some_name'
        self.assertRaises(AssertionError, lambda: func.variables["some_name"])
        # But can use function argument 'my_arg'
        ref = func.variables["my_arg"]
        self.assertCodeEqual(ref.as_source_code(), "my_arg")

    def test_function_args_name_check(self):
        module = codegen.Module()
        module.reserve_name("my_arg")
        func_name = module.reserve_name("myfunc")
        self.assertRaises(
            AssertionError,
            codegen.Function,
            func_name,
            args=["my_arg"],
            parent_scope=module,
        )

    def test_function_args_name_reserved_check(self):
        module = codegen.Module()
        module.reserve_function_arg_name("myarg")
        func_name = module.reserve_name("myfunc")
        func = codegen.Function(func_name, args=["myarg"], parent_scope=module)
        func.body.value = func.variables["myarg"]
        func = codegen.simplify(func)
        self.assertCodeEqual(
            func.as_source_code(),
            """
           myfunc myarg =
               myarg
        """,
        )

    def test_add_assignment_reserved(self):
        scope = codegen.Scope()
        let = codegen.Let(parent_scope=scope)
        name = let.add_assignment("x", codegen.String("a string"))
        self.assertEqual(name.name, "x")
        let.value = codegen.String("other")
        self.assertCodeEqual(
            let.as_source_code(),
            """
            let
                x = "a string"
            in
                "other"
        """,
        )

    def test_function_call_args(self):
        scope = codegen.Scope()
        scope.reserve_name(
            "aFunction", type=types.Function(dtypes.Number, dtypes.String)
        )
        func_call = codegen.FunctionCall(
            scope.variables["aFunction"], [codegen.Number(123)]
        )
        self.assertCodeEqual(func_call.as_source_code(), "aFunction 123")

    def test_function_call_using_apply(self):
        scope = codegen.Scope()
        scope.reserve_name(
            "aFunction", type=types.Function(dtypes.Number, dtypes.String)
        )
        func_call = scope.variables["aFunction"].apply(codegen.Number(123))
        self.assertCodeEqual(func_call.as_source_code(), "aFunction 123")

    def test_function_call_nested(self):
        scope = codegen.Scope()
        scope.reserve_name(
            "aFunction", type=types.Function(dtypes.Number, dtypes.String)
        )
        scope.reserve_name(
            "aFunction2", type=types.Function(dtypes.String, dtypes.Number)
        )
        func_call_1 = scope.variables["aFunction"].apply(codegen.Number(123))
        func_call_2 = scope.variables["aFunction2"].apply(func_call_1)
        self.assertCodeEqual(func_call_2.as_source_code(), "aFunction2 (aFunction 123)")

    def test_if(self):
        scope = codegen.Scope()
        if_expr = codegen.If(parent_scope=scope)
        if_expr.condition = codegen.Equals(codegen.Number(1), codegen.Number(2))
        if_expr.true_branch.value = codegen.Number(3)
        if_expr.false_branch.value = codegen.Number(4)
        if_expr = codegen.simplify(if_expr)
        self.assertCodeEqual(
            if_expr.as_source_code(),
            """
            if (1 == 2) then
                3
            else
                4
        """,
        )

    def test_string_join_empty(self):
        join = codegen.StringConcat([])
        join = codegen.simplify(join)
        self.assertCodeEqual(join.as_source_code(), '""')

    def test_string_join_one(self):
        join = codegen.StringConcat([codegen.String("hello")])
        join = codegen.simplify(join)
        self.assertCodeEqual(join.as_source_code(), '"hello"')

    def test_string_join_two(self):
        scope = codegen.Scope()
        scope.reserve_name("tmp", type=dtypes.String)
        var = scope.variables["tmp"]
        join = codegen.StringConcat([codegen.String("hello "), var])
        self.assertCodeEqual(
            join.as_source_code(),
            """
            String.concat [ "hello "
                          , tmp
                          ]
        """,
        )

    def test_string_join_collapse_strings(self):
        scope = codegen.Scope()
        scope.reserve_name("tmp", type=dtypes.String)
        var = scope.variables["tmp"]
        join1 = codegen.StringConcat(
            [
                codegen.String("hello "),
                codegen.String("there "),
                var,
                codegen.String(" how"),
                codegen.String(" are you?"),
            ]
        )
        join1 = codegen.simplify(join1)
        self.assertCodeEqual(
            join1.as_source_code(),
            """
            String.concat [ "hello there "
                          , tmp
                          , " how are you?"
                          ]
            """,
        )

    def test_string_join_omit_empty(self):
        scope = codegen.Scope()
        scope.reserve_name("tmp", type=dtypes.String)
        var = scope.variables["tmp"]
        join = codegen.StringConcat([codegen.String(""), var, var])
        join = codegen.simplify(join)
        self.assertCodeEqual(
            join.as_source_code(),
            """
            String.concat [ tmp
                          , tmp
                          ]
        """,
        )

    def test_string_join_omit_empty_2(self):
        scope = codegen.Scope()
        scope.reserve_name("tmp", type=dtypes.String)
        var = scope.variables["tmp"]
        join = codegen.StringConcat([codegen.String(""), var])
        join = codegen.simplify(join)
        self.assertCodeEqual(
            join.as_source_code(),
            """
            tmp
        """,
        )

    def test_cleanup_name(self):
        for n, c in [
            ("abc-def()[]ghi,.<>¡!?¿", "abcdefghi"),  # illegal chars
            ("1abc", "n1abc"),  # leading digit not allowed
            ("-", "n"),  # aboid being empty after removing illegals
            ("_abc", "n_abc"),  # leading underscore not allowed
            ("abc_def", "abc_def"),  # underscore in middle is allowed
        ]:
            self.assertEqual(codegen.cleanup_name(n), c)

    def test_case(self):
        scope = codegen.Scope()
        tmp = scope.reserve_name("tmp")
        case = codegen.Case(scope.variables[tmp])
        branch1 = case.add_branch(codegen.String("x"))
        branch1.value = codegen.Number(1)
        branch2 = case.add_branch(codegen.String("y"))
        branch2.value = codegen.Number(2)
        branch3 = case.add_branch(codegen.Otherwise())
        branch3.value = codegen.Number(3)

        case = codegen.simplify(case)
        self.assertCodeEqual(
            case.as_source_code(),
            """
            case tmp of
                "x" ->
                    1
                "y" ->
                    2
                _ ->
                    3
        """,
        )

    def test_record_update(self):
        rec = types.Record()
        rec.add_field("name", dtypes.String)
        rec.add_field("age", dtypes.Number)
        rec.add_field("height", dtypes.Number)
        scope = codegen.Scope()
        tmp = scope.reserve_name("tmp", type=rec)
        var = scope.variables[tmp]
        update = codegen.RecordUpdate(
            var, name=codegen.String("Fred"), age=codegen.Number(34)
        )
        self.assertCodeEqual(
            update.as_source_code(),
            """
            { tmp | age = 34, name = "Fred" }
        """,
        )

    def test_multiple_indent(self):
        scope = codegen.Scope()
        scope.reserve_name(
            "aFunction",
            type=types.Function.for_multiple_inputs(
                [dtypes.Number, dtypes.Number], dtypes.String
            ),
        )
        let1 = codegen.Let()
        name1 = let1.add_assignment("x", codegen.Number(1))
        name2 = let1.add_assignment("y", codegen.Number(2))
        let1.value = codegen.Add(name1, name2)

        case1 = codegen.Case(codegen.Add(codegen.Number(7), codegen.Number(8)))
        branch1 = case1.add_branch(codegen.Number(15))
        branch1.value = codegen.Number(16)
        branch2 = case1.add_branch(codegen.Otherwise())
        branch2.value = codegen.Number(17)

        let2 = codegen.Let()
        name3 = let2.add_assignment("a", codegen.Number(3))
        name4 = let2.add_assignment("b", case1)
        let2.add_assignment("c", codegen.Number(4))
        let2.value = codegen.Add(name3, name4)

        func_call = scope.variables["aFunction"].apply(let1, let2)
        func_call = codegen.simplify(func_call)
        self.assertCodeEqual(
            func_call.as_source_code(),
            """
        aFunction (let
                       x = 1
                       y = 2
                   in
                       (x + y)) (let
                                     a = 3
                                     b = case (7 + 8) of
                                             15 ->
                                                 16
                                             _ ->
                                                 17

                                     c = 4
                                 in
                                     (a + b))
        """,
        )



from __future__ import absolute_import, unicode_literals

from collections import OrderedDict, defaultdict
from functools import wraps

import attr
import six

from elm_fluent import exceptions

# This module is the heart of the type tracking the compiler does. Types are
# associated with node objects in the codegen.py. It is currently rather adhoc
# and incomplete, but suffices for our purposes.
#
# There are 4 main purposes of this type tracking:
#
# 1. To prevent (or provide early feedback) for many type errors in the development
#    of the compiler, so that generated code has a much better chance of at least
#    being type correct.
#
# 2. To catch many type errors the users (i.e. FTL authors) might make, for example
#    if a message has both `DATETIME($arg)` and `NUMBER($arg)` it is a clear type error
#
# 3. For tracing the types we need to work out what conversions need to be done
#    at which stage e.g. numbers need to be run through NumberFormat.format
#    eventually if they need to be converted to strings, but not eagerly because
#    select expressions need to handle them as numbers.
#
# 4. For determining the correct types of the passed in record argument, so that we can
#    generate correct type signatures for the generated functions.
#
# However, for item 1, while the current code worked OK initially, after the
# addition of type parameters there are now many bugs e.g. an expression like
# `Maybe.Just.apply(codegen.Number(1))` should have a type of 'Maybe Int', but
# actually returns 'Maybe a', all TypeParams are considered equal to each other
# etc.
#
# These do not currently affect the correctness of output. To fix these issues,
# and others marked as TODO in this module, it would probably be best to start
# over with this whole type system, using proper Hindler-Milney type inference
# perhaps. See
# https://github.com/rob-smallshire/hindley-milner-python/blob/master/inference.py


def with_auto_env(meth):
    @wraps(meth)
    def wrapper(self, from_module, env=None):
        if env is None:
            env = SignatureEnv()
        return meth(self, from_module, env=env)

    return wrapper


@attr.s
class SignatureEnv(object):
    """
    Environment object for creating type/function signatures
    """

    used_type_variables = attr.ib(factory=dict)


class ElmType(object):
    def constrain(self, other):
        """
        Constrain the other type to be compatible with (equal to or more specific
        than) self May return a new object, other unchanged, or other mutated.
        (TODO - would be cleaner if it never mutated, only returned new
        objects).

        """
        raise NotImplementedError(
            "{0} needs to implement 'constrain'".format(self.__class__)
        )

    def apply_args(self, args):
        if len(args) > 0:
            raise AssertionError(
                "{0} is not a function, cannot apply arguments to it".format(self)
            )
        return self

    def signature_sub_types(self):
        """
        Returns all the type objects that would appear in a signature
        """
        raise NotImplementedError(
            "{0} needs to implement signature_sub_types".format(self.__class__)
        )


class UnconstrainedType(ElmType):
    def constrain(self, other):
        return other

    @with_auto_env
    def as_signature(self, from_module, env=None):
        # This fails if we reach 'z', which is unlikely
        if not env.used_type_variables:
            retval = "a"
        else:
            retval = six.unichr(max(map(ord, env.used_type_variables)) + 1)
        env.used_type_variables[retval] = self
        return retval

    def signature_sub_types(self):
        return []


class TypeParam(UnconstrainedType):
    def __init__(self, preferred_name):
        self.preferred_name = preferred_name

    def constrain(self, other):
        if isinstance(other, UnconstrainedType):
            return self
        else:
            return other

    @with_auto_env
    def as_signature(self, from_module, env=None):
        for v, t in env.used_type_variables.items():
            if t is self:
                return v
        c = 1
        candidate = self.preferred_name
        while candidate in env.used_type_variables:
            c = c + 1
            candidate = "{0}{1}".format(self.preferred_name, c)
        env.used_type_variables[candidate] = self
        return candidate

    def __eq__(self, other):
        # This is a really weak check, but deliberately relaxes things enough
        # so we can easily get signatures to work the way we need
        return isinstance(other, TypeParam)


class Type(ElmType):
    def __init__(
        self, full_name, module, params=None, constructors=None, reserve_names=True
    ):
        """
        Construct a union type.
        constructors is a list of constructors, where each constructor is a:
           string - for nullary constructors
           tuple of (string, types...) - for constructors with type parameters
        """

        from . import codegen

        if " " in full_name:
            assert (
                params is None
            ), "Either pass params, or auto created using spaces in name"
            parts = full_name.split(" ")
            name, param_names = parts[0], tuple(parts[1:])
            params = [TypeParam(p) for p in param_names]
        else:
            name = full_name
            if params is None:
                params = []
        self.name = name
        param_dict = OrderedDict()
        for p in params:
            param_dict[p.preferred_name] = p
        self.param_dict = param_dict
        self.module = module
        if constructors is None:
            constructors = []

        for c in constructors:
            if isinstance(c, six.text_type):
                name = c
                params = ()
                constructor_type = self
            else:
                name, params = c[0], tuple(c[1:])
                params = [
                    self.param_dict[p] if isinstance(p, six.text_type) else p
                    for p in params
                ]
                constructor_type = Function.for_multiple_inputs(params, self)

            if reserve_names:
                reserved_name = module.reserve_name(name, type=constructor_type)
                assert reserved_name == name, "Expected {0} == {1}".format(
                    reserved_name, name
                )

            if module.is_default_imports:
                # For builtins, we attach a name to self for easy access.
                attr_name = name
                # For 'True', 'False' etc, we have to avoid a clash with Python keywords.
                # Could use module 'keyword.kwlist', but this is different on Python 2.
                if attr_name in ["True", "False"]:
                    attr_name += "_"

                setattr(self, attr_name, codegen.VariableReference(name, module))

    def __eq__(self, other):
        return self._is_compatible(other) and (self.param_dict == other.param_dict)

    def _is_compatible(self, other):
        return (
            isinstance(other, Type)
            and (self.module == other.module)
            and (self.name == other.name)
        )

    def __str__(self):
        return self.as_signature(dummy_module)

    def __repr__(self):
        return "<ElmType: {0}>".format(str(self))

    @with_auto_env
    def as_signature(self, from_module, env=None):
        return "{0}{1}".format(
            from_module.get_name_qualifier(self.module), self.name
        ) + "".join(
            " " + type_paren_wrap(t.as_signature(from_module, env=env))
            for n, t in self.param_dict.items()
        )

    def signature_sub_types(self):
        return self.param_dict.values()

    def constrain(self, other):
        if isinstance(other, UnconstrainedType):
            return self
        if not self._is_compatible(other):
            raise exceptions.TypeMismatch(
                "{0} is not compatible with {1}".format(self, other)
            )
        diff_params = {
            n: p
            for n, p in self.param_dict.items()
            if other.param_dict[n] != self.param_dict[n]
        }
        if diff_params:
            retval = other.specialize(**diff_params)
        else:
            retval = self
        return retval

    def specialize(self, **params):
        retval = self.clone()
        for name, type_obj in params.items():
            if name not in retval.param_dict:
                raise LookupError(
                    "{0} is not a parameter of type {1}".format(name, self)
                )
            retval.param_dict[name] = type_obj.constrain(self.param_dict[name])
        return retval

    def clone(self):
        retval = self.__class__(
            self.name, self.module, params=None, constructors=None, reserve_names=False
        )
        retval.param_dict.update(self.param_dict)
        return retval


class Tuple(Type):
    def __init__(self, *param_types):
        type_params = [
            TypeParam(six.unichr(ord("a") + i)) for i in range(len(param_types))
        ]
        super(Tuple, self).__init__(
            "Tuple", None, params=type_params, reserve_names=False
        )
        for type_param, param_type in zip(type_params, param_types):
            self.param_dict[type_param.preferred_name] = param_type

    def clone(self):
        retval = self.__class__(self.name)
        retval.param_dict.update(self.param_dict)
        return retval

    @with_auto_env
    def as_signature(self, from_module, env=None):
        return "({0})".format(
            ", ".join(
                t.as_signature(from_module, env=env) for n, t in self.param_dict.items()
            )
        )


def type_paren_wrap(sig):
    if " " in sig and not (sig.startswith("(") and sig.endswith(")")):
        return "({0})".format(sig)
    else:
        return sig


@attr.s
class TypeSource(object):
    ftl_source = attr.ib()
    type_obj = attr.ib()


class Record(ElmType):
    def __init__(self, **fields):
        """
        Pass fields as keyword args to create a fixed record type,
        or no parameters to create one that is extensible
        """
        self.fields = {}  # Dict from name to type
        self.fixed = False
        for name, type_obj in fields.items():
            self.add_field(name, type_obj)
        self.fixed = bool(fields)
        if not self.fixed:
            self.field_type_ftl_sources = defaultdict(list)

    def add_field(self, name, type_obj=None, from_ftl_source=None):
        """
        Add a field, or set the type for an existing field.
        """
        if type_obj is None:
            type_obj = UnconstrainedType()
        if name in self.fields:
            if self.fixed:
                assert type_obj.constrain(self.fields[name])
            else:
                if from_ftl_source is not None and type_obj != self.fields[name]:
                    self.field_type_ftl_sources[name].append(
                        TypeSource(from_ftl_source, type_obj)
                    )
                try:
                    new_type = type_obj.constrain(self.fields[name])
                except exceptions.TypeMismatch as e:
                    raise exceptions.RecordTypeMismatch(
                        *e.args, record_type=self, field_name=name
                    )
                self.fields[name] = new_type
        else:
            if self.fixed:
                raise AssertionError(
                    "Cannot add field {0} to a fixed record type, only {1} available".format(
                        name, ", ".join(self.fields.keys())
                    )
                )
            self.fields[name] = type_obj
            if from_ftl_source is not None:
                self.field_type_ftl_sources[name].append(
                    TypeSource(from_ftl_source, type_obj)
                )

    def constrain(self, other):
        if isinstance(other, UnconstrainedType):
            return self

        if not isinstance(other, Record):
            raise exceptions.TypeMismatch(
                "Expected type {0} is not compatible with a record type".format(other)
            )

        if other.fixed:
            for n in self.fields:
                if n not in other.fields:
                    raise exceptions.TypeMismatch(
                        "Unexpected field {0}, only {1} available".format(
                            n, ", ".join(self.other.keys())
                        )
                    )
        else:
            # TODO - it would be cleaner if we were returning a clone here,
            # but it gets really tricky to re-associate the new object as part
            # of the type of the codegen.Function objects create.
            for n, t in self.fields.items():
                other.field_type_ftl_sources[n].extend(self.field_type_ftl_sources[n])
                other.add_field(n, t)

        return other

    @with_auto_env
    def as_signature(self, from_module, env=None):
        def fields_signature():
            return ", ".join(
                "{0} : {1}".format(name, type_obj.as_signature(from_module, env=env))
                for name, type_obj in sorted(self.fields.items())
            )

        if self.fixed:
            return "{ %s }" % fields_signature()
        else:
            base = UnconstrainedType().as_signature(from_module, env=env)
            if not self.fields:
                return base
            else:
                return "{ %s | %s }" % (base, fields_signature())

    def signature_sub_types(self):
        return self.fields.values()


class Function(ElmType):
    def __init__(self, input_type, output_type):
        # In Elm, functions always have a single input, and a single output. We
        # have conveniences for 'multiple input' functions later.
        self.input_type = input_type
        self.output_type = output_type

    def __repr__(self):
        return "<Function: {0}>".format(self.as_signature(dummy_module))

    def __eq__(self, other):
        return (
            isinstance(other, Function)
            and (self.input_type == other.input_type)
            and (self.output_type == other.output_type)
        )

    def constrain(self, other):
        assert isinstance(other, Function), "Expecting {0} to be a Function".format(
            other
        )
        return Function(
            other.input_type.constrain(self.input_type),
            other.output_type.constrain(self.output_type),
        )

    @staticmethod
    def for_multiple_inputs(input_types, output_type):
        if len(input_types) == 0:
            return output_type
        if len(input_types) == 1:
            return Function(input_types[0], output_type)
        else:
            return Function(
                input_types[0],
                Function.for_multiple_inputs(input_types[1:], output_type),
            )

    @with_auto_env
    def as_signature(self, from_module, env=None):
        return "{0} -> {1}".format(
            self.input_type.as_signature(from_module, env=env),
            self.output_type.as_signature(from_module, env=env),
        )

    def signature_sub_types(self):
        return [self.input_type, self.output_type]

    def apply_args(self, args, from_ftl_source=None):
        """
        Returns that type that would remain after the supplied arguments
        (which are expression objects) are applied.

        Also applies type constraints to the args.
        """
        if len(args) == 0:
            return self
        arg, remainder = args[0], args[1:]

        # 1. Suppose we have a function
        #  { a | foo : String } -> String
        # and an empty record type {}.
        # When we apply the function to the type,
        # the record type ought to gain a 'foo : String' field

        # 2. Suppose we have a function
        #  List a -> a
        # and an input
        #  List String
        #
        # When we apply the function to the type,
        # the output type should end up as String.
        #
        # TODO fix this second case somehow???

        arg.constrain_type(self.input_type, from_ftl_source=from_ftl_source)
        return self.output_type.apply_args(remainder)


class DummyModule(object):
    def get_name_qualifier(self, module):
        return ""


dummy_module = DummyModule()


def signature_traverse(type_obj):
    sub_parts = type_obj.signature_sub_types()
    for part in sub_parts:
        for t in signature_traverse(part):
            yield t
    yield type_obj



# -*- coding: utf-8 -*-

from __future__ import absolute_import, unicode_literals

import os
import subprocess
import sys
import time
import unittest

from click.testing import CliRunner
from pyvirtualdisplay import Display
from selenium import webdriver

from elm_fluent import cli

THIS_DIR = os.path.dirname(os.path.abspath(__file__))
TEST_PROJECT = os.path.join(THIS_DIR, "test_project")


def noisy_check_call(cmd):
    sys.stdout.write("\n" + " ".join(cmd) + "\n")
    subprocess.check_call(cmd)


@unittest.skipIf(os.environ.get("TEST_FAST_RUN", "0") == "1", "Skipping slow tests")
class TestEndToEnd(unittest.TestCase):

    visible = os.environ.get("TEST_SHOW_BROWSER", "0") == "1"
    use_network = os.environ.get("TEST_NO_NETWORK", "0") != "1"

    @classmethod
    def setUpClass(cls):
        super(TestEndToEnd, cls).setUpClass()
        if not cls.visible:
            display_args = {"visible": False}
            cls.__display = Display(**display_args)
            cls.__display.start()
        cls.browser = webdriver.Firefox()

    @classmethod
    def tearDownClass(cls):
        cls.browser.quit()
        if not cls.visible:
            cls.__display.stop()
        super(TestEndToEnd, cls).tearDownClass()

    def test_everything(self):
        os.chdir(TEST_PROJECT)
        # Run in process, for the sake of coverage testng
        runner = CliRunner()
        result = runner.invoke(
            cli.main,
            ["--when-missing=fallback", "--no-bdi-isolating"],
            catch_exceptions=False,
        )
        self.assertEqual(result.exit_code, 0)

        # Check the script in installed as expected
        noisy_check_call(["ftl2elm", "--when-missing=fallback", "--no-bdi-isolating"])

        if self.use_network:
            noisy_check_call(["elm-install"])
        elm_make_cmd = ["elm-make", "--yes", "Main.elm", "--output=main.js"]
        if "TRAVIS_BUILD_DIR" in os.environ:
            # See https://github.com/elm/compiler/issues/1473#issuecomment-245704142
            elm_make_cmd = [
                os.path.join(
                    os.environ["TRAVIS_BUILD_DIR"], "sysconfcpus", "bin", "sysconfcpus"
                ),
                "-n",
                "2",
            ] + elm_make_cmd

        noisy_check_call(elm_make_cmd)
        self.browser.get("file://{0}/main.html".format(TEST_PROJECT))
        page_source = self.browser.page_source

        # Static tests
        self.assertIn("This is the title", page_source)
        self.assertIn("This is a test page for elm-fluent", page_source)

        # Number
        self.assertIn("There are 12,345 things", page_source)
        self.assertIn("There are 123,456.7 things", page_source)
        self.assertIn("There are 4,567.8 things", page_source)
        self.assertIn("There are 7890 things", page_source)
        self.assertIn("You have 6,543.21 US dollars in your bank.", page_source)

        # Select
        self.assertIn("You've clicked 0 times", page_source)
        e = self.browser.find_element_by_css_selector("button#increment")
        e.click()
        time.sleep(0.1)
        page_source = self.browser.page_source
        self.assertIn("You've clicked once", page_source)
        e.click()
        time.sleep(0.1)
        page_source = self.browser.page_source
        self.assertIn("You've clicked 2 times", page_source)

        # Select with mixed numerics
        self.assertIn("You have no new messages", page_source)
        self.assertIn("You have one new message", page_source)
        self.assertIn("You have 2 new messages", page_source)

        # Parameterized terms
        self.assertIn("We have the parameterized things", page_source)

        # Dates:
        self.assertIn("January 1, 1970 AD, 01:02:03", page_source)

        # Html:
        self.assertIn("Some text with this &amp; that", page_source)
        self.assertIn(
            "Some <b>bold text</b> and some <b>bold <i>and italic</i></b> text.",
            page_source,
        )
        self.assertIn('Some <span class="foo">highlighted text</span>.', page_source)
        self.assertIn("Hello <b>Mary</b>!", page_source)
        self.assertIn("Hello <b>Maria</b>! You came back", page_source)
        self.assertIn('<b foo="Hello">Some text</b>', page_source)

        # Dynamic HTML:
        self.assertIn("You haven't moved yet", page_source)
        self.assertIn(
            '<a data-left="" class="direction" href="#">left</a>', page_source
        )
        self.assertIn(
            '<a data-right="" class="direction" href="#">right</a>', page_source
        )

        e = self.browser.find_element_by_css_selector("a[data-left]")
        e.click()
        time.sleep(0.1)
        page_source = self.browser.page_source
        self.assertIn("You moved left", page_source)

        e = self.browser.find_element_by_css_selector("a[data-right]")
        e.click()
        time.sleep(0.1)
        page_source = self.browser.page_source
        self.assertIn("You moved right", page_source)



"""
Stub to define types for the Intl.PluralRules module
"""
from __future__ import absolute_import, unicode_literals

from .. import codegen, types
from . import defaults as dtypes, intl_locale

module = codegen.Module(name="Intl.PluralRules")

PluralRules = types.Type("PluralRules", module)

module.reserve_name("fromLocale", type=types.Function(intl_locale.Locale, PluralRules))

module.reserve_name(
    "select",
    type=types.Function.for_multiple_inputs(
        [PluralRules, dtypes.Number], dtypes.String
    ),
)



"""
Stub to define types for the Intl.TimeZone module
"""
from __future__ import absolute_import, unicode_literals

from .. import codegen, types
from . import defaults as dtypes

module = codegen.Module(name="Intl.TimeZone")


TimeZone = types.Type("TimeZone", module, constructors=[("TimeZone", dtypes.String)])



"""
Stub to define types for the Fluent module
"""
from __future__ import absolute_import, unicode_literals

from .. import codegen, types
from . import (
    date,
    defaults as dtypes,
    html,
    intl_datetimeformat,
    intl_locale,
    intl_numberformat,
)

module = codegen.Module(name="Fluent")

# Numbers

FluentNumber = types.Type("FluentNumber number", module)

module.reserve_name("number", type=types.Function(dtypes.Number, FluentNumber))

module.reserve_name(
    "formattedNumber",
    type=types.Function.for_multiple_inputs(
        [intl_numberformat.Options, dtypes.Number], FluentNumber
    ),
)

module.reserve_name(
    "reformattedNumber",
    type=types.Function.for_multiple_inputs(
        [intl_numberformat.Options, FluentNumber], FluentNumber
    ),
)

module.reserve_name(
    "numberFormattingOptions",
    type=types.Function(FluentNumber, intl_numberformat.Options),
)

module.reserve_name(
    "formatNumber",
    type=types.Function.for_multiple_inputs(
        [intl_locale.Locale, FluentNumber], dtypes.String
    ),
)

module.reserve_name("numberValue", type=types.Function(FluentNumber, dtypes.Number))

# Dates

FluentDate = types.Type("FluentDate", module)

module.reserve_name("date", type=types.Function(date.Date, FluentDate))

module.reserve_name(
    "formattedDate",
    type=types.Function.for_multiple_inputs(
        [intl_datetimeformat.Options, date.Date], FluentDate
    ),
)

module.reserve_name(
    "reformattedDate",
    type=types.Function.for_multiple_inputs(
        [intl_datetimeformat.Options, FluentDate], FluentDate
    ),
)

module.reserve_name(
    "dateFormattingOptions",
    type=types.Function(FluentDate, intl_datetimeformat.Options),
)

module.reserve_name(
    "formatDate",
    type=types.Function.for_multiple_inputs(
        [intl_locale.Locale, FluentDate], dtypes.String
    ),
)


module.reserve_name(
    "selectAttributes",
    type=types.Function.for_multiple_inputs(
        [
            dtypes.List.specialize(
                a=types.Tuple(dtypes.String, dtypes.List.specialize(a=html.Attribute))
            ),
            dtypes.List.specialize(a=dtypes.String),
        ],
        dtypes.List.specialize(a=html.Attribute),
    ),
)



# -*- coding: utf-8 -*-

from __future__ import absolute_import, unicode_literals

import unittest

from elm_fluent import codegen, types
from elm_fluent.stubs import defaults as dtypes


class TestTypes(unittest.TestCase):
    def test_elm_type_str(self):
        self.assertEqual(str(dtypes.String), "String")

    def test_elm_type_signature(self):
        self.assertEqual(dtypes.String.as_signature(codegen.Module()), "String")

    def test_elm_type_eq(self):
        self.assertEqual(dtypes.String, dtypes.String)
        self.assertNotEqual(dtypes.String, dtypes.Number)

    def test_function_signature(self):
        self.assertEqual(
            types.Function(dtypes.String, dtypes.Number).as_signature(codegen.Module()),
            "String -> number",
        )

        self.assertEqual(
            types.Function.for_multiple_inputs(
                [dtypes.String, dtypes.Number], dtypes.Bool
            ).as_signature(codegen.Module()),
            "String -> number -> Bool",
        )

    def test_function_eq(self):
        function1 = types.Function.for_multiple_inputs(
            [dtypes.String, dtypes.Number], dtypes.String
        )
        function2 = types.Function.for_multiple_inputs(
            [dtypes.String, dtypes.Number], dtypes.String
        )
        function3 = types.Function.for_multiple_inputs(
            [dtypes.Bool, dtypes.Number], dtypes.String
        )

        self.assertEqual(function1, function2)
        self.assertNotEqual(function1, function3)

    def test_function_apply_args(self):
        def typed_expr(type_obj):
            class Expr(codegen.Expression):
                type = type_obj

                def constrain_type(self, type_obj, from_ftl_source=None):
                    self.type = type_obj

            return Expr()

        function = types.Function.for_multiple_inputs(
            [dtypes.String, dtypes.Number], dtypes.Bool
        )

        zero_applied = function.apply_args([])
        self.assertEqual(
            zero_applied.as_signature(codegen.Module()), "String -> number -> Bool"
        )
        self.assertEqual(zero_applied, function)

        one_applied = function.apply_args([typed_expr(dtypes.String)])
        self.assertEqual(one_applied.as_signature(codegen.Module()), "number -> Bool")
        self.assertEqual(one_applied, types.Function(dtypes.Number, dtypes.Bool))

        two_applied = function.apply_args(
            [typed_expr(dtypes.String), typed_expr(dtypes.Number)]
        )
        self.assertEqual(two_applied.as_signature(codegen.Module()), "Bool")
        self.assertEqual(two_applied, dtypes.Bool)

    def test_unconstrained_signature(self):
        u = types.UnconstrainedType()
        self.assertEqual(u.as_signature(codegen.Module()), "a")

    def test_function_signature_with_type_variables(self):
        function = types.Function.for_multiple_inputs(
            [types.UnconstrainedType(), types.UnconstrainedType()],
            types.UnconstrainedType(),
        )
        self.assertEqual(function.as_signature(codegen.Module()), "a -> b -> c")

    def test_empty_record_signature(self):
        self.assertEqual(types.Record().as_signature(codegen.Module()), "a")

    def test_empty_record_signature_in_function(self):
        function = types.Function(types.Record(), types.UnconstrainedType())
        self.assertEqual(function.as_signature(codegen.Module()), "a -> b")

    def test_one_field_record_signature(self):
        r = types.Record()
        r.add_field("foo", type_obj=dtypes.String)
        self.assertEqual(r.as_signature(codegen.Module()), "{ a | foo : String }")

    def test_two_field_record_signature(self):
        r = types.Record()
        r.add_field("foo", type_obj=dtypes.String)
        r.add_field("bar")
        self.assertEqual(
            r.as_signature(codegen.Module()), "{ a | bar : b, foo : String }"
        )

    def test_one_field_fixed_record_signature(self):
        r = types.Record(foo=dtypes.String)
        self.assertEqual(r.as_signature(codegen.Module()), "{ foo : String }")

    def test_type_variables_complex(self):
        r1 = types.Record()
        r1.add_field("foo")
        r1.add_field("bar")
        r2 = types.Record()
        r2.add_field("baz")
        function = types.Function.for_multiple_inputs(
            [r1, r2], types.UnconstrainedType()
        )
        self.assertEqual(
            function.as_signature(codegen.Module()),
            "{ a | bar : b, foo : c } -> { d | baz : e } -> f",
        )

    def test_non_builtin_simple_types(self):
        source_module = codegen.Module(name="MyModule")
        mytype = types.Type("MyType", source_module)
        self.assertEqual(mytype.as_signature(source_module), "MyType")

        main_module = codegen.Module()
        main_module.add_import(source_module, "MyAlias")
        self.assertEqual(mytype.as_signature(main_module), "MyAlias.MyType")

    def test_type_parameters_signatures(self):
        source_module = codegen.Module(name="MyModule")
        dict_type = types.Type("Dict k v", source_module)
        self.assertEqual(dict_type.as_signature(source_module), "Dict k v")
        env = types.SignatureEnv(used_type_variables={"k": 1, "v": 2})
        self.assertEqual(dict_type.as_signature(source_module, env=env), "Dict k2 v2")
        dict_type = types.Type("Dict k v", source_module)
        str_to_float_dict = dict_type.specialize(k=dtypes.String, v=dtypes.Float)
        self.assertEqual(
            str_to_float_dict.as_signature(source_module), "Dict String Float"
        )

        container_type = types.Type("Container a", source_module)
        complex_type = container_type.specialize(a=dict_type)
        self.assertEqual(
            complex_type.as_signature(source_module), "Container (Dict k v)"
        )

    def test_type_parameters_constructors(self):
        source_module = codegen.Module(name="MyModule")
        container_type = types.Type(
            "Container a", source_module, constructors=["Empty", ("Single", "a")]
        )
        self.assertEqual(container_type.as_signature(source_module), "Container a")
        Empty = source_module.variables["Empty"]
        Single = source_module.variables["Single"]
        self.assertEqual(Empty.type.as_signature(source_module), "Container a")
        self.assertEqual(Single.type.as_signature(source_module), "a -> Container a")

    def test_function_signature_type_parameters(self):
        param = types.TypeParam("a")
        f = types.Function(param, param)
        self.assertEqual(f.as_signature(codegen.Module()), "a -> a")

        f2 = types.Function(param, dtypes.List.specialize(a=param))
        self.assertEqual(f2.as_signature(codegen.Module()), "a -> List a")

    def test_tuple_signature(self):
        module = codegen.Module(name="MyModule")
        t = types.Tuple(dtypes.String, dtypes.Number)
        self.assertEqual(t.as_signature(module), "(String, number)")

    # This, or something like it, would require a better type system
    #
    # def test_type_parameters_function_application(self):
    #     source_module = codegen.Module(name="MyModule")
    #     container_type = types.Type(
    #         "Container a", source_module, constructors=["Empty", ("Single", "a")]
    #     )
    #     Single = source_module.variables["Single"]

    #     # a -> Container a
    #     function_1 = Single.type
    #     applied_1 = function_1.apply_args([codegen.Number(1)])
    #     self.assertEqual(applied_1, container_type.specialize(a=dtypes.Int))

    #     # Container a -> String -> Container a
    #     function_2 = types.Function.for_multiple_inputs(
    #         [container_type, dtypes.String], container_type
    #     )
    #     applied_2 = function_2.apply_args([codegen.String("")])
    #     self.assertEqual(applied_2, container_type.specialize(a=dtypes.String))

    #     # Container a -> String -> a
    #     function_3 = types.Function.for_multiple_inputs(
    #         [container_type, dtypes.String], container_type.param_dict["a"]
    #     )
    #     applied_3 = function_3.apply_args([codegen.String("")])
    #     self.assertEqual(applied_3, dtypes.String)

    def test_type_parameters_constrain(self):
        source_module = codegen.Module(name="MyModule")
        container_type = types.Type("Container a", source_module)
        specialized = container_type.specialize(a=dtypes.Int)
        specialized2 = container_type.specialize(a=dtypes.Int)
        self.assertEqual(specialized, specialized2)
        self.assertNotEqual(specialized, container_type)
        self.assertEqual(specialized.constrain(container_type), specialized)



# -*- coding: utf-8 -*-
"""Console script for elm_fluent."""
from __future__ import absolute_import, unicode_literals

import os.path
import time

import attr
import click
import watchdog.events
import watchdog.observers

from . import __version__
from .run import ErrorWhenMissing, FallbackToDefaultLocaleWhenMissing, run_compile


@attr.s
class CompilationOptions(object):
    locales_dir = attr.ib()
    output_dir = attr.ib()
    default_locale = attr.ib()
    missing_translation_strategy = attr.ib()
    use_isolating = attr.ib()
    cwd = attr.ib()
    verbose = attr.ib(default=False)


@click.command()
@click.option(
    "--locales-dir",
    default="locales",
    help="Location of the locales directory that holds all FTL files.",
)
@click.option("--output-dir", default=".", help="Location of the outputted Elm files.")
@click.option(
    "--when-missing",
    default="error",
    type=click.Choice(["error", "fallback"]),
    help="What to do when translations are missing for a locale, defaults to error",
)
@click.option(
    "--default-locale", default="en", help="The default locale, used for fallbacks"
)
@click.option(
    "--bdi-isolating/--no-bdi-isolating",
    default=True,
    help="Use BDI isolating characters",
)
@click.option(
    "--watch",
    "watch",
    flag_value=True,
    help="Watch for changes and rebuild as necessary",
)
@click.option("--verbose/--quiet", default=False, help="More verbose output")
@click.option("--version", "version", flag_value=True, help="Print version and exit")
def main(
    locales_dir,
    output_dir,
    when_missing,
    default_locale,
    bdi_isolating,
    watch,
    verbose,
    version,
):
    if version:
        click.echo("elm-fluent {0}".format(__version__))
        return

    locales_dir = os.path.normpath(os.path.abspath(locales_dir))
    if not os.path.exists(locales_dir) or not os.path.isdir(locales_dir):
        raise click.UsageError(
            "Locales directory '{0}' does not exist. Please specify a correct locales "
            "directory using the --locales-dir option".format(locales_dir)
        )

    output_dir = os.path.normpath(os.path.abspath(output_dir))
    if not os.path.exists(output_dir) or not os.path.isdir(output_dir):
        raise click.UsageError(
            "Output directory '{0}' does not exist. Please specify a correct output "
            "directory using the --output-dir option".format(output_dir)
        )

    if when_missing == "error":
        missing_translation_strategy = ErrorWhenMissing()
    elif when_missing == "fallback":
        missing_translation_strategy = FallbackToDefaultLocaleWhenMissing(
            default_locale
        )

    options = CompilationOptions(
        locales_dir=locales_dir,
        output_dir=output_dir,
        default_locale=default_locale,
        missing_translation_strategy=missing_translation_strategy,
        use_isolating=bdi_isolating,
        cwd=os.getcwd(),
        verbose=verbose,
    )

    if watch:
        run_compile_and_ignore_abort(options)
        observer = watchdog.observers.Observer()
        handler = RunCompileEventHandler(options)
        observer.schedule(handler, options.locales_dir, recursive=True)
        observer.start()
        try:
            while True:
                time.sleep(1)
        except KeyboardInterrupt:
            observer.stop()
        observer.join()
    else:
        return run_compile(options)


def run_compile_and_ignore_abort(options):
    try:
        run_compile(options)
    except click.Abort:
        pass


class RunCompileEventHandler(watchdog.events.FileSystemEventHandler):
    def __init__(self, options):
        self.options = options

    def on_any_event(self, event):
        click.secho("Changes detected, re-running...\n", fg="yellow")
        run_compile_and_ignore_abort(self.options)



"""
Stub to define types for the Date module
"""
from __future__ import absolute_import, unicode_literals

from .. import codegen, types

# https://package.elm-lang.org/packages/elm-lang/core/latest/Date
module = codegen.Module(name="Date")
Date = types.Type("Date", module)



from __future__ import absolute_import, print_function, unicode_literals

import os
import os.path

import click
import language_tags

from . import exceptions
from .compiler import (
    compile_master,
    compile_messages,
    module_name_for_locale,
    span_to_position,
)
from .stubs import defaults as dtypes


class MissingTranslationStrategy(object):
    def get_locale_when_missing(self, locale):
        raise NotImplementedError()

    def missing_ftl_file(self, options, filename, locale, errors, warnings):
        raise NotImplementedError()

    def missing_message(self, message_id, locale, errors, warnings):
        raise NotImplementedError()


def missing_file(filename, options):
    return exceptions.MissingMessageFile(
        "Message file '{0}' not found".format(os.path.relpath(filename, options.cwd))
    )


class ErrorWhenMissing(MissingTranslationStrategy):
    def get_locale_when_missing(self, locale):
        return None

    def missing_ftl_file(self, options, filename, locale, errors, warnings):
        errors.append(missing_file(filename, options))

    def missing_message(self, message_id, locale, errors, warnings):
        errors.append(
            exceptions.MissingMessage(
                "Locale '{0}' - Message '{1}' missing".format(locale, message_id)
            )
        )


class FallbackToDefaultLocaleWhenMissing(MissingTranslationStrategy):
    def __init__(self, fallback_locale):
        self.fallback_locale = fallback_locale

    def _can_fallback(self, locale):
        return locale != self.fallback_locale

    def get_locale_when_missing(self, locale):
        if self._can_fallback(locale):
            return self.fallback_locale
        else:
            return None

    def missing_ftl_file(self, options, filename, locale, errors, warnings):
        if not self._can_fallback(locale):
            # Can't fall back to anything if the fallback is missing
            errors.append(missing_file(filename, options))
        else:
            warnings.append(missing_file(filename, options))

    def missing_message(self, message_id, locale, errors, warnings):
        if locale == self.fallback_locale:
            extra = (
                " This message will be omitted since it is not in the fallback locale."
            )
        else:
            extra = ""
        warnings.append(
            exceptions.MissingMessage(
                "Locale '{0}' - Message '{1}' missing.{2}".format(
                    locale, message_id, extra
                )
            )
        )


def run_compile(options):
    locales = find_locales(options.locales_dir)
    if not locales:
        raise click.UsageError(
            "No locale directories (directories containing .ftl files) found in {0} directory".format(
                options.locales_dir
            )
        )
    bad_locales = [l for l in locales if not language_tags.tags.check(l)]
    if bad_locales:
        raise click.UsageError(
            "The following directory names are not valid BCP 47 language tags: {0}".format(
                ", ".join(bad_locales)
            )
        )

    stems = find_all_ftl_stems(options.locales_dir, locales)
    finalizers = []
    error_printers = []
    warning_printers = []
    for stem in stems:
        success, finalizer, error_printer, warning_printer = generate_elm_for_stem(
            options, locales, stem
        )
        if success:
            finalizers.append(finalizer)
        if error_printer:
            error_printers.append(error_printer)
        if warning_printer:
            warning_printers.append(warning_printer)

    if warning_printers:
        click.secho("\nWarnings:\n", fg="red", bold=True)
        for wp in warning_printers:
            wp()

    if error_printers:
        click.secho("\nErrors:\n", fg="red", bold=True)
        for ep in error_printers:
            ep()
        if options.verbose:
            click.secho("Failed!", fg="red", bold=True)
        raise click.Abort()

    if options.verbose:
        click.secho("\nWriting files:\n", fg="green", bold=True)
    for f in finalizers:
        f()

    if options.verbose:
        click.secho("Success!", fg="green", bold=True)


def generate_elm_for_stem(options, locales, stem):
    errors = []
    warnings = []

    sources = {}
    locale_modules = {}
    modules_to_write = []
    master_message_mapping = {}
    for locale in locales:
        filename = os.path.join(options.locales_dir, locale, stem)
        module_name = module_name_for_stem(stem, locale=locale)
        if os.path.exists(filename):
            with open(filename, "rb") as f:
                messages_string = f.read().decode("utf-8")
            sources[filename] = messages_string
            module, compile_errors, message_mapping = compile_messages(
                messages_string,
                locale,
                message_source=filename,
                module_name=module_name,
                use_isolating=options.use_isolating,
            )
            locale_modules[locale] = module
            master_message_mapping.update(message_mapping)
            new_elm_path = path_for_module(options, module_name)
            if compile_errors:
                errors.extend(compile_errors)
            else:
                modules_to_write.append((new_elm_path, module))
        else:
            options.missing_translation_strategy.missing_ftl_file(
                options, filename, locale, errors, warnings
            )

    master_module_name = module_name_for_stem(stem, master=True)
    try:
        master_module, master_errors, master_warnings = compile_master(
            master_module_name, locales, locale_modules, master_message_mapping, options
        )
    except Exception as e:
        click.secho(
            "While compiling {0}, an exception occurred.".format(master_module_name),
            fg="red",
            bold=True,
        )
        click.secho(
            "Please report this as a bug to https://github.com/elm-fluent/elm-fluent",
            fg="red",
            bold=True,
        )
        raise e

    errors.extend(master_errors)
    warnings.extend(master_warnings)

    if not master_errors:
        master_filename = path_for_module(options, master_module_name)
        modules_to_write.append((master_filename, master_module))

    def finalizer():
        if not errors:
            for fname, module in modules_to_write:
                if module.exports:
                    # If there are no exports, there is no point writing the
                    # module, and `exposing ()` is also invalid syntax, so we
                    # must avoid writing it
                    source = module.as_source_code()
                    ensure_path_dirs(fname)
                    if options.verbose:
                        click.echo(
                            "Writing {0}".format(os.path.relpath(fname, options.cwd))
                        )
                    with open(fname, "wb") as f:
                        f.write(source.encode("utf-8"))

    def error_printer():
        print_errors(options, errors, sources)

    def warning_printer():
        print_warnings(warnings)

    return (
        len(errors) == 0,
        finalizer,
        error_printer if errors else None,
        warning_printer if warnings else None,
    )


def print_errors(options, errors, sources):
    for err in errors:
        if getattr(err, "error_sources", []):
            for error_source in err.error_sources:
                source_filename = error_source.message_source
                row, col = span_to_position(
                    error_source.expr.span, sources[source_filename]
                )
                short_filename = os.path.relpath(source_filename, options.cwd)
                if error_source.message_id is not None:
                    click.echo(
                        "{0}:{1}:{2}: In message '{3}': {4}".format(
                            short_filename,
                            row,
                            col,
                            error_source.message_id,
                            err.args[0],
                        )
                    )
                else:
                    click.echo(
                        "{0}:{1}:{2}: {3}".format(short_filename, row, col, err.args[0])
                    )
        else:
            if hasattr(err, "message_func_name"):
                click.echo(
                    "While trying to compile master '{0}' function:".format(
                        err.message_func_name
                    )
                )
                click.echo("  {0}".format(err.args[0]))
            else:
                click.echo(err.args[0])
        if isinstance(err, exceptions.RecordTypeMismatch):
            click.echo(
                "  Explanation: incompatible types were detected for message argument '${0}'".format(
                    err.field_name
                )
            )
            type_sources = err.record_type.field_type_ftl_sources[err.field_name]
            if type_sources:
                click.echo("  Compare the following:")
                for type_source in type_sources:
                    t_source_filename = type_source.ftl_source.message_source
                    t_short_filename = os.path.relpath(t_source_filename, options.cwd)
                    row, col = span_to_position(
                        type_source.ftl_source.expr.span, sources[t_source_filename]
                    )
                    click.echo(
                        "    {0}:{1}:{2}: Inferred type: {3}".format(
                            t_short_filename, row, col, type_source.type_obj
                        )
                    )

                if any(
                    type_source.type_obj == dtypes.String
                    for type_source in type_sources
                ):
                    click.echo("")
                    click.echo(
                        "  Hint: You may need to use NUMBER() or DATETIME() builtins to force the correct type"
                    )


def print_warnings(warnings):
    for warning in warnings:
        click.echo(warning.args[0])


def ensure_path_dirs(path):
    dirname = os.path.dirname(path)
    if not os.path.exists(dirname):
        os.makedirs(dirname)


def module_name_for_stem(ftl_stem, locale=None, master=False):
    if locale is None:
        assert master
    if locale is not None:
        assert not master

    if master:
        first_part = "Translations"
    else:
        first_part = module_name_for_locale(locale)

    return "Ftl.{0}.{1}".format(
        first_part,
        ".".join(part.title() for part in ftl_stem.replace(".ftl", "").split("/")),
    )


def path_for_module(options, module_name):
    return os.path.join(options.output_dir, module_name.replace(".", "/") + ".elm")


def find_locales(locales_dir):
    return [
        d for d in os.listdir(locales_dir) if contains_ftl(os.path.join(locales_dir, d))
    ]


def contains_ftl(directory):
    if not os.path.isdir(directory):
        return False
    return any(
        is_ftl(f) for dirpath, dirnames, files in os.walk(directory) for f in files
    )


def is_ftl(path):
    return path.endswith(".ftl") and not os.path.basename(path).startswith(".")


def find_all_ftl_stems(locales_dir, locales):
    """
    Given a locales directory and a list of locales, finds all the
    ftl stem names. For example, for these files:

    /path/to/locales/en/foo/bar.ftl
    /path/to/locales/de/foo/bar.ftl

    there is a single ftl stem, 'foo/bar.ftl'
    """
    ftl_stems = set([])
    for l in locales:
        locale_base_dir = os.path.join(locales_dir, l)
        stem_offset = len(locale_base_dir) + 1
        ftl_files = [
            os.path.join(dirpath, f)[stem_offset:]
            for dirpath, dirnames, files in os.walk(locale_base_dir)
            for f in files
            if is_ftl(f)
        ]
        ftl_stems |= set(ftl_files)
    return sorted(list(ftl_stems))



"""
Types for Elm defaults
"""
from __future__ import absolute_import, unicode_literals

from elm_fluent import codegen, types

# Default imports: see http://package.elm-lang.org/packages/elm-lang/core/latest
# or https://github.com/elm-lang/core/blob/5.1.1/src/Basics.elm

# As we use things from this list, we add types for them (below) and comment
# them out here
ELM_DEFAULT_IMPORTS = set(
    [
        "max",
        "min",
        "Order",
        "LT",
        "EQ",
        "GT",
        "compare",
        "not",
        "&&",
        "||",
        "xor",
        "+",
        "-",
        "*",
        "/",
        "^",
        "//",
        "rem",
        "%",
        "negate",
        "abs",
        "sqrt",
        "clamp",
        "logBase",
        "e",
        "pi",
        "cos",
        "sin",
        "tan",
        "acos",
        "asin",
        "atan",
        "atan2",
        "round",
        "floor",
        "ceiling",
        "truncate",
        "toFloat",
        "degrees",
        "radians",
        "turns",
        "toPolar",
        "fromPolar",
        "isNaN",
        "isInfinite",
        "toString",
        "++",
        "identity",
        "always",
        "<|",
        "|>",
        "<<",
        ">>",
        "flip",
        "curry",
        "uncurry",
        "Never",
        "never",
        # "List",
        "::",
        # "Maybe", "Just", "Nothing",  - added below
        "Result",
        "Ok",
        "Err",
        "String",
        "Tuple",
        "Debug",
        "Program",
        "Cmd",
        "!",
        "Sub",
    ]
)


class DefaultImports(codegen.Scope):
    is_default_imports = True


default_imports = DefaultImports()

for i in ELM_DEFAULT_IMPORTS:
    default_imports.reserve_name(i)


String = types.Type("String", default_imports)

Float = types.Type("Float", default_imports)

Int = types.Type("Int", default_imports)

# 'number' is a hard-coded Elm 'type class', we can treat it as a simple type
# for our purposes
Number = types.Type("number", default_imports)


Bool = types.Type("Bool", default_imports, constructors=["True", "False"])


Maybe = types.Type(
    "Maybe a", default_imports, constructors=["Nothing", ("Just", types.TypeParam("a"))]
)

List = types.Type("List a", default_imports)
